/* ExpPyRE.c -- support for perl-compatible regular expressions */
/* Copyright (C) 1998 Michael P. Reilly, All rights reserved
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Log: ExpPyTcl.c,v $
 * Revision 1.15  1999/07/04 22:05:07  arcege
 * Add code for Tcl 8.1 later
 *
 * Revision 1.14  1999/04/25 17:04:19  arcege
 * Fix the logic of when to free; this might cause core dumps
 *
 * Revision 1.13  1999/04/25 15:52:49  arcege
 * Wrong cast inside Python type declaration
 *
 * Revision 1.12  1999/04/20 18:07:51  arcege
 * Fix CVS log string;  use new common malloc interface
 *
 *
 */

#include "ExpPycnf.h"
#include <Python.h>
#include <tcl.h>
#include <expect.h>

#include "ExpPyMem.h"
#include "ExpPyTcl.h"

/* exceptions */
static PyObject *ExpPyTcl_Error;

static char version[] =
 "1.1.4";
static char copyright[] =
  "Copyright (C) 1998 Michael P. Reilly, All rights reserved";

typedef struct {
  PyObject_HEAD
#ifdef TCL_8_1
  TCL_REGEXP *re;
#else
  struct regexp *re;
#endif
  char *string;
  PyObject *patt;
} ExpPyTcl_regexp;

staticforward PyTypeObject ExpPyTclRE_Type;
#define is_ExpPyTclRE(v)        ((v)->ob_type == &ExpPyTclRE_Type)
#define ExpPyTclRE_re(v)        (((ExpPyTcl_regexp *)(v))->re)
#define ExpPyTclRE_str(v)       (((ExpPyTcl_regexp *)(v))->string)
#define ExpPyTclRE_patt(v)      (((ExpPyTcl_regexp *)(v))->patt)

/* accessor functions (for use in expect.c) */
int
ExpPyTclRE_Check(self)
  PyObject *self;
  {
    return is_ExpPyTclRE(self);
  }
TCL_REGEXP *
ExpPyTclRE_regexp(self)
  PyObject *self;
  {
    return (regexp *)ExpPyTclRE_re(self);
  }
char *
ExpPyTclRE_pattern(self)
  PyObject *self;
  {
#if PYTHON_API_VERSION >= 1007
    return PyString_AS_STRING(ExpPyTclRE_patt(self));
#else
    return PyString_AsString(ExpPyTclRE_patt(self));
#endif
  }

static void
ExpPyTclRE_dealloc(self)
  PyObject *self;
  {
    if (ExpPyTclRE_re(self))
      free(ExpPyTclRE_re(self));
    if (ExpPyTclRE_str(self))
      ExpPy_free(ExpPyTclRE_str(self));

    Py_XDECREF(ExpPyTclRE_patt(self));
    PyMem_DEL(self);
  }

#define group_set(r, p) ((r)->startp[p] != NULL && (r)->endp[p] != NULL)
#define group_value(s, type, p) (ExpPyTclRE_re(s)->type[p] - ExpPyTclRE_str(s))

static PyObject *
ExpPyTclRE_group(self, args)
  PyObject *self, *args;
  { int i, size, pos, start, end;
    PyObject *t, *item, *returnVal = NULL;
    TCL_REGEXP *re = ExpPyTclRE_re(self);

    PyErr_Clear();
    if ((size = PyTuple_Size(args)) == 1) {
      if (PyArg_ParseTuple(args, "i", &pos) && pos < NSUBEXP) {
        if (group_set(re, pos)) {
          start = group_value(self, startp, pos);
          end   = group_value(self, endp, pos);
          returnVal = Py_BuildValue("(ii)", start, end);
        } else {
          Py_INCREF(Py_None);
          returnVal = Py_None;
        }
      } else if (pos >= NSUBEXP) {
        PyErr_SetString(PyExc_IndexError, "index out of range");
      }

    } else {
      if ((returnVal = PyTuple_New(size)) != NULL) {
        for (i = 0; i < size && !PyErr_Occurred(); i++) {
          if ((item = PyTuple_GetItem(args, i)) != NULL &&
              PyArg_Parse(item, "i", &pos) &&
              (pos < NSUBEXP)) {
            if (!group_set(re, pos))
              t = Py_BuildValue("(OO)", Py_None, Py_None);
            else {
              start = group_value(self, startp, pos);
              end   = group_value(self, endp, pos);
              t = Py_BuildValue("(ii)", start, end);
            }
            PyTuple_SetItem(returnVal, i, t);
          } else if (pos >= NSUBEXP)
            PyErr_SetString(PyExc_IndexError, "index out of range");
        }

        if (PyErr_Occurred()) {
          Py_DECREF(returnVal);
          returnVal = NULL;
        }
      } else
        PyErr_NoMemory();
    }
    return returnVal;
  }

static PyObject *
ExpPyTclRE_match(self, args)
  PyObject *self, *args;
  { PyObject *returnVal = NULL;
    char *pattern;
    int i;

    if (PyArg_ParseTuple(args, "s", &pattern)) {
      if (ExpPyTclRE_str(self) != NULL)
        ExpPy_free(ExpPyTclRE_str(self));
      ExpPyTclRE_str(self) = ExpPy_calloc(char, strlen(pattern));
      strcpy(ExpPyTclRE_str(self), pattern);
      i = TclRegExec(
            ExpPyTclRE_re(self),
            ExpPyTclRE_str(self),
            ExpPyTclRE_str(self));
      returnVal = Py_BuildValue("i", i);
    }
    return returnVal;
  }

static struct PyMethodDef ExpPyTclRE_methods[] = {
  { "group",        ExpPyTclRE_group,        METH_VARARGS },
  { "match",        ExpPyTclRE_match,        METH_VARARGS },
  { NULL,           NULL }
};

static PyObject *
ExpPyTclRE_getattr(self, name)
  PyObject *self;
  char *name;
  { PyObject *returnVal;

    if (strcmp(name, "pattern") == 0) {
      Py_INCREF(ExpPyTclRE_patt(self));
      returnVal = ExpPyTclRE_patt(self);
    } else if (strcmp(name, "__members__") == 0) {
      returnVal = Py_BuildValue("[s]",
        "pattern");
    } else {
      returnVal = Py_FindMethod(ExpPyTclRE_methods, (PyObject *)self, name);
    }
    return returnVal;
  }

static int
ExpPyTclRE_length(self)
  PyObject *self;
  { int i;
    for (i = 0; i < NSUBEXP; i++)
      if (!group_set(ExpPyTclRE_re(self), i))
        break;
    return i;
  }

static PyObject *
ExpPyTclRE_getitem(self, position)
  PyObject *self;
  int position;
  { int size;
    PyObject *returnVal = NULL;
    TCL_REGEXP *re = ExpPyTclRE_re(self);

    if (position >= NSUBEXP) {
      PyErr_SetString(PyExc_IndexError, "index out of range");
    } else if (!group_set(re, position)) {
      Py_INCREF(Py_None);
      returnVal = Py_None;
    } else {
      size = re->endp[position] - re->startp[position];
      returnVal = PyString_FromStringAndSize(re->startp[position], size);
    }
    return returnVal;
  }

#define bounds(v,l) (((v) > (l)) ? (l) : \
  (((v) < -(l)) ? 0 : \
  (((v) < 0) ? ((l) + (v)) : \
  (v))))

static PyObject *
ExpPyTclRE_getslice(self, low, high)
  PyObject *self;
  int low, high;
  { int num, size, i;
    PyObject *returnVal;
    TCL_REGEXP *re = ExpPyTclRE_re(self);

    num = ExpPyTclRE_length(self);
    high = bounds(high, num);
    low = bounds(low, num);
    if (low >= high)
      returnVal = PyList_New(0);
    else {
      returnVal = PyList_New(high - low);

      for (i=0; low < high; i++, low++) {
        size = re->endp[i] - re->startp[i];
        PyList_SetItem(returnVal, i,
          PyString_FromStringAndSize(re->startp[i], size));
      }
    }
    return returnVal;
  }
#undef bounds

static PySequenceMethods ExpPyTclRE_as_sequence = {
  (inquiry)ExpPyTclRE_length,              /*sq_length*/
  (binaryfunc)0,                           /*sq_concat*/
  (intargfunc)0,                           /*sq_repeat*/
  (intargfunc)ExpPyTclRE_getitem,          /*sq_item*/
  (intintargfunc)ExpPyTclRE_getslice,      /*sq_slice*/
  (intobjargproc)0,                        /*sq_ass_item*/
  (intintobjargproc)0,                     /*sq_ass_slice*/
};

static PyObject *
ExpPyTclRE_repr(self)
  PyObject *self;
  { PyObject *repr_patt, *string, *result = NULL;
    char buffer[256];

    string = repr_patt = NULL;
    if ((string = PyObject_Repr(ExpPyTclRE_patt(self))) &&
        (repr_patt = PySequence_GetSlice(string, 1, -1))) { /* remove quotes */
      strcpy(buffer, "<Tclregexp /");
#if PYTHON_API_VERSION >= 1007
      strcat(buffer, PyString_AS_STRING(repr_patt)); /*self->patt));*/
#else
      strcat(buffer, PyString_AsString(repr_patt)); /*self->patt));*/
#endif
      strcat(buffer, "/>");
      result = PyString_FromString(buffer);
    }
    Py_XDECREF(string);
    Py_XDECREF(repr_patt);
    return result;
  }

statichere PyTypeObject ExpPyTclRE_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "Tclregexp",
  sizeof(ExpPyTclRE_Type),
  0,
  (destructor)ExpPyTclRE_dealloc,    /*tp_dealloc*/
  (printfunc)0,                      /*tp_print*/
  (getattrfunc)ExpPyTclRE_getattr,   /*tp_getattr*/
  (setattrfunc)0,                    /*tp_setattr*/
  (cmpfunc)0,                        /*tp_compare*/
  (reprfunc)ExpPyTclRE_repr,         /*tp_repr*/
  0,                                 /*tp_as_number*/
  &ExpPyTclRE_as_sequence,           /*tp_as_sequence*/
  0,                                 /*tp_as_maggin*/
  (hashfunc)0,                       /*tp_hash*/
  (ternaryfunc)0,                    /*tp_call*/
  (reprfunc)0,                       /*tp_str*/
};

PyObject *
ExpPyTclRE_new(self, args)
  PyObject *self, *args;
  { ExpPyTcl_regexp *new = NULL;
    PyObject *obj;
    char *pattern;
    int i;

    if (PyArg_ParseTuple(args, "S", &obj) &&
        (new = PyObject_NEW(ExpPyTcl_regexp, &ExpPyTclRE_Type)) != NULL) {
      ExpPyTclRE_re(new) = NULL;
      Py_INCREF(obj);
      ExpPyTclRE_patt(new) = (PyObject *)obj;
#if PYTHON_API_VERSION >= 1007
      pattern = PyString_AS_STRING(obj);
#else
      pattern = PyString_AsString(obj);
#endif
      if ((ExpPyTclRE_re(new) = TclRegComp(pattern)) == NULL) {
        ExpPyTclRE_dealloc(new);
        PyErr_SetString(ExpPyTcl_Error, TclGetRegError());
        new = NULL;
      } else {
        for (i=0; i < NSUBEXP; i++)
          ExpPyTclRE_re(new)->startp[i] =
            ExpPyTclRE_re(new)->endp[i] = NULL;
      }
      ExpPyTclRE_str(new) = NULL;
    }
    return (PyObject *)new;
  }

static PyMethodDef ExpPyTcl_methods[] = {
  { "compile", ExpPyTclRE_new, METH_VARARGS},
  { NULL, NULL }
};

void
initExpPyTcl()
  { PyObject *mod, *dict;
    PyObject *constant;

    mod = Py_InitModule("tclre", ExpPyTcl_methods);
    dict = PyModule_GetDict(mod);

    ExpPyTcl_Error = PyString_FromString("tclre.error");
    PyDict_SetItemString(dict, "error", ExpPyTcl_Error);
    constant = PyString_FromString(version);
    PyDict_SetItemString(dict, "__version__", constant);
    constant = PyString_FromString(copyright);
    PyDict_SetItemString(dict, "__copyright__", constant);
  }

