/*  ExpPyint.c  -- implimentation of the interact function */
/* Copyright (C) 1999 Michael P. Reilly, All rights reserved
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Log: ExpPyint.c,v $
 * Revision 1.16  1999/12/25 16:51:15  arcege
 * Add buffer segments to the Pbuffer object
 *
 * Revision 1.15  1999/12/25 12:10:20  arcege
 * Add assertion function
 *
 * Revision 1.14  1999/12/24 13:25:02  arcege
 * Change the mapping object into a PyObject (tuple).  Change buffer into
 * an object (which can handle more complicated operations).  Add Python-
 * thread safety code.  Lots of (too many?) minor changes.
 *
 * Revision 1.13  1999/12/21 13:58:52  arcege
 * Fix where we re-acquire Python interp; move get_fileno() to a better position
 *
 * Revision 1.12  1999/07/19 12:31:02  arcege
 * Add some initial form of interrupt sequence to interact()
 *
 * Revision 1.11  1999/07/05 14:40:28  arcege
 * Add helper macros; when not including methods, raise an exception
 *
 * Revision 1.10  1999/07/05 11:57:38  arcege
 * Make code more compilable (K&R)
 *
 * Revision 1.9  1999/07/04 15:04:37  arcege
 * Linux (and maybe some other systems) give an error on read of a terminal
 * when EOF is detected; make the interact() method return as EOF.
 *
 * Revision 1.8  1999/04/25 15:34:50  arcege
 * Add doc string for interact function.  Convert stream structure into
 * a Python object (for future use).
 *
 * Revision 1.7  1999/04/23 21:36:48  arcege
 * Fix more EOF issues with interact
 *
 * Revision 1.6  1999/04/23 12:08:45  arcege
 * Raise TimeoutError inside interact function, as appropriate
 *
 * Revision 1.5  1999/04/23 11:40:18  arcege
 * Allow None to be given as an argument to the interact() function; denotes
 * that input or output should be ignored.  Use bit fields instead of int's
 * in the stream structure.  Fix some of the logic within the reader and
 * writer functions.  Fix the logic for what exception is raised (this was
 * causing core dumps).
 *
 * Revision 1.4  1999/04/20 18:06:20  arcege
 * Add copyright information;  use new common malloc interface;  make the
 * pyexp_interact function Python-thread safe.
 *
 */

#include <Python.h>
#include <stdio.h>
#include <expect.h>
#include <errno.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#include "ExpPyMem.h"

static PyObject *
AssertionError(text)
  char *text;
  {
#if PYTHON_API_VERSION >= 1007
    PyErr_SetString(PyExc_AssertionError, text);
#else
    PyErr_SetString(PyExc_SystemError, text);
#endif
    return NULL;
  }

#define STRM_BUFLEN 256
#define NO_STREAM_NUM -2

#define is_null(s) ((s) == NO_STREAM_NUM)

/* convert a Python object into a file descriptor */
static int
get_fileno(file)
  PyObject *file;
  { int fd = -1;
    PyObject *tmp, *method;

    if (PyInt_Check(file))
#if PYTHON_API_VERSION >= 1007
      fd = PyInt_AS_LONG(file);
#else
      fd = PyInt_AsLong(file);
#endif
    else if (file == Py_None)
      fd = NO_STREAM_NUM;
    else {
      if ((method = PyObject_GetAttrString(file, "fileno")) != NULL)
        if (PyCallable_Check(method)) {
          if ((tmp = PyObject_CallObject(method, NULL)) != NULL &&
              PyInt_Check(tmp))
#if PYTHON_API_VERSION >= 1007
            fd = PyInt_AS_LONG(tmp);
#else
            fd = PyInt_AsLong(tmp);
#endif
          else
            PyErr_SetString(PyExc_TypeError,
              "fileno method returned a non-integer");
          Py_XDECREF(tmp);
        } else
          PyErr_SetString(PyExc_TypeError, "call of non-function");
      else {
#if 0  /* if int(object) could be taken as file descriptor */
        if ((tmp = PyNumber_Int(file)) != NULL) {
#if PYTHON_API_VERSION >= 1007
          fd = PyInt_AS_LONG(tmp);
#else
          fd = PyInt_AsLong(tmp);
#endif
          Py_XDECREF(tmp);
        } else
#endif
          PyErr_SetString(PyExc_TypeError, "call of non-function");
      }
      Py_XDECREF(method);
    }
    return fd;
  }

staticforward PyTypeObject Pbuffer_Type;

struct buffer_seg {
  int len;
  char *p;
  struct buffer_seg *next;
};

typedef struct Pbuffer {
  PyObject_HEAD
  int len;
  int nseg;
  struct buffer_seg *n;
} Pbuffer;

#define Pbuffer_Check(v)   ((v)->ob_type == &Pbuffer_Type)
#define Pbuffer_nseg(v)    (((Pbuffer *)(v))->nseg)
#define Pbuffer_len(v)     (((Pbuffer *)(v))->len)
#define Pbuffer_seg(v)     (((Pbuffer *)(v))->n)

static void
clear_bufseg(seg, l, h)
  struct buffer_seg *seg;
  int l, h;
  { int i;

    for (i = l; i < h; i++)
      seg->p[i] = '\000';
  }

static void
write_to_bufseg(seg, l, s)
  struct buffer_seg *seg;
  int l;
  char *s;
  { int i;

    if (l <= seg->len) {
      for (i = 0; i < l; i++)
        seg->p[i] = s[i];
    }
  }

static struct buffer_seg *
make_bufseg(length)
  int length;
  { struct buffer_seg *new;

    new = ExpPy_malloc(struct buffer_seg);
    if (new != NULL) {
      new->len = length;
      new->next = NULL;
      if (length <= 0)
        new->p = NULL;
      else if ((new->p = ExpPy_calloc(char, length)) != NULL)
        clear_bufseg(new, 0, length);
      else {
        ExpPy_free(new);
        new = NULL;
      }
    }
    if (new == NULL)
      PyErr_NoMemory();
    return new;
  }

static struct buffer_seg *
destroy_bufseg(seg)
  struct buffer_seg *seg;
  { struct buffer_seg *next;

    if (seg != NULL) {
      next = seg->next;
      if (seg->p != NULL)
        ExpPy_free(seg->p);
      ExpPy_free(seg);
    } else
      next = NULL;
    return next;
  }

static int
destroy_bufseg_chain(seg, seg_until)
  struct buffer_seg *seg, *seg_until;
  { int i = 0;

    while (seg != seg_until && seg != NULL) {
      seg = destroy_bufseg(seg);
      i++;
    }
    return i;
  }

static struct buffer_seg *
copy_bufseg(seg)
  struct buffer_seg *seg;
  { struct buffer_seg *new = NULL;
    int i;

    if (seg != NULL &&
        (new = make_bufseg(seg->len)) != NULL) {
      write_to_bufseg(new, seg->len, seg->p);
      new->next = copy_bufseg(seg->next);
    }
    return new;
  }

static struct buffer_seg *
break_bufseg_at(seg, pos)
  struct buffer_seg *seg;
  int pos;
  { struct buffer_seg *new;
    int i, len;

    len = seg->len - pos;
    new = make_bufseg(len);
    if (new != NULL) {
      write_to_bufseg(new, len, &seg->p[pos]);
      clear_bufseg(seg, pos, seg->len);
      seg->len = pos;
      new->next = seg->next;
      seg->next = new;
    }
    return new;
  }

static struct buffer_seg *
join_bufseg(seg, len)
  struct buffer_seg *seg;
  int len;
  { struct buffer_seg *new;
    int i, j;

    if (seg->next == NULL && seg->len == len)
      return seg;
    new = make_bufseg(len);
    j = 0;
    while (seg != NULL && j < len) {
      i = 0;
      while (i < seg->len)
        new->p[j++] = seg->p[i++];
      seg = destroy_bufseg(seg);
    }
    return new;
  }

static PyObject *
Pbuffer_dump(self, stream)
  PyObject *self;
  FILE *stream;
  { PyObject *obj, *t, *s;
    struct buffer_seg *seg;
    int nseg;

    obj = PyTuple_New(2);
    PyTuple_SetItem(obj, 0, PyString_FromString("Pbuffer"));
    for ( nseg = 0, seg = Pbuffer_seg(self);
          seg != NULL;
          nseg++, seg = seg->next )
      ;
    t = PyTuple_New(nseg);
    PyTuple_SetItem(obj, 1, t);
    for ( nseg = 0, seg = Pbuffer_seg(self);
          seg != NULL;
          nseg++, seg = seg->next ) {
      s = Py_BuildValue("(is#)", nseg, seg->p, seg->len);
      PyTuple_SetItem(t, nseg, s);
    }
    if (stream != NULL)
      PyObject_Print(obj, stream, 0);
    return obj;
  }

/* remove the empty segments from the front */
static void
Pbuffer_normalize(self)
  PyObject *self;
  { int notdone = 1;
    while (notdone)
      if (Pbuffer_seg(self)->len == 0)
        Pbuffer_seg(self) = destroy_bufseg(Pbuffer_seg(self));
      else
        notdone = 0;
    if (Pbuffer_seg(self) == NULL)
      Pbuffer_seg(self) = make_bufseg(0);
  }

/* return the segment and index (in that segment) of the "global" position
   in the buffer.
 */
static int
Pbuffer_position(self, pos, seg)
  PyObject *self;
  int pos;
  struct buffer_seg **seg;
  { struct buffer_seg *s;
    int i;

    if (seg == NULL)
      return -2;
    if (pos < 0)
      return -1;
    i = 0;
    s = Pbuffer_seg(self);
    while (s != NULL && (i + s->len) < pos) {
      i += s->len;
      s = s->next;
    }
    if (s == NULL)
      return -1;
    *seg = s;
    return pos - i;
  }

static PyObject *
Pbuffer_NEW(str, len)
  char *str;
  int len;
  { Pbuffer *object;
    struct buffer_seg *b;

    if (len <= 0)
      len = 0;
    object = PyObject_NEW(Pbuffer, &Pbuffer_Type);
    if (object != NULL) {
      Pbuffer_len(object) = len;
      if ((b = make_bufseg(len)) == NULL) {
        PyMem_DEL(object);
        return NULL;
      }
      Pbuffer_nseg(object) = 1;
      Pbuffer_seg(object) = b;
      if (str != NULL)
        write_to_bufseg(b, len, str);
    }
    return (PyObject *)object;
  }

static void
Pbuffer__dealloc(self)
  Pbuffer *self;
  { struct buffer_seg *n;
    
    n = Pbuffer_seg(self);
    while (n != NULL)
      n = destroy_bufseg(n);
    PyMem_DEL(self);
  }

static int
Pbuffer__length(self)
  PyObject *self;
  {
    return Pbuffer_len(self);
  }

static PyObject *
Pbuffer__concat(self, other)
  PyObject *self, *other;
  { PyObject *result = NULL;
    struct buffer_seg *seg;
    int sl, ol;

    sl = Pbuffer_len(self);
    ol = Pbuffer_len(other);
    result = Pbuffer_NEW(NULL, sl + ol);
    if (result != NULL && Pbuffer_len(result) > 0) {
      destroy_bufseg(Pbuffer_seg(result));
      seg = Pbuffer_seg(result) = copy_bufseg(Pbuffer_seg(self));
      while (seg->next != NULL)
        seg = seg->next;
      seg->next = copy_bufseg(Pbuffer_seg(other));
    }
    return result;
  }

static PyObject *
Pbuffer__repeat(self, count)
  PyObject *self;
  int count;
  { PyObject *result = NULL;
    struct buffer_seg *next, *new;

    if ((result = Pbuffer_NEW(NULL, 0)) != NULL)
      if (count > 0) {
        Pbuffer_nseg(result) = count * Pbuffer_nseg(self);
        Pbuffer_len(result)  = count * Pbuffer_len(self);
        next = NULL;
        while (--count > 0) {
          if ((new = copy_bufseg(Pbuffer_seg(self))) == NULL) {
            Py_DECREF(result);
            result = NULL;
            break;
          }
          /* tack on to end of segment list */
          if (next == NULL) {
            destroy_bufseg(Pbuffer_seg(result));
            Pbuffer_seg(self) = new;
          } else
            next->next = new;
          /* advance to the end of the current copy */
          next = new;
          while (next->next != NULL)
            next = next->next;
        }
      }
    return result;
  }

static PyObject *
Pbuffer__getitem(self, index)
  PyObject *self;
  int index;
  { PyObject *result = NULL;
    struct buffer_seg *seg;
    int i;

    if (index < 0)
      index += Pbuffer_len(self);
    if (0 <= index && index < Pbuffer_len(self)) {
      i = Pbuffer_position(self, index, &seg);
      if (i < 0)
        result = AssertionError("index out of range");
      else
        result = PyString_FromStringAndSize(&seg->p[i], 1);
    } else {
      PyErr_SetString(PyExc_IndexError, "index out of range");
    }
    return result;
  }

static PyObject *
Pbuffer__getslice(self, low, high)
  PyObject *self;
  int low, high;
  { PyObject *result = NULL;
    struct buffer_seg *seg;
    char *s;
    int len, i, j;

    if (low < 0)
      low += Pbuffer_len(self);
    if (high < 0)
      high += Pbuffer_len(self);
    if (low > high || low >= Pbuffer_len(self) || high < 0) {
      len = 0;
      high = low = 0;
    } else
      len = high - low;
    if (len <= 0)
      result = Pbuffer_NEW(NULL, 0);
    else {
      i = Pbuffer_position(self, low, &seg);
      if (i < 0)
        result = AssertionError("index out of range");
      else {
        s = ExpPy_calloc(char, len+1);
        high -= (low + i);
        j = 0;
        while (seg != NULL && i < high) {
          if (i >= seg->len) {
            high -= seg->len;
            i = 0;
            seg = seg->next;
          } else
            s[j++] = seg->p[i++];
        }
        s[j] = '\000';
        result = Pbuffer_NEW(s, len);
      }
    }
    return result;
  }

static int
Pbuffer__delitem(self, index)
  PyObject *self;
  int index;
  { int result = -1;
    struct buffer_seg *seg;
    int i;

    if (0 < index && index < Pbuffer_len(self)) {
      i = Pbuffer_position(self, index, &seg);
      if (i < 0)
        (void)AssertionError("index out of range");
      else {
        while (i < seg->len-1) {
          seg->p[i] = seg->p[i+1];
          i++;
        }
        seg->p[i] = '\000';
        seg->len--;
        Pbuffer_len(self)--;
        result = 0;
      }
    } else
      PyErr_SetString(PyExc_IndexError, "index out of range");
    return result;
  }

static int
Pbuffer__setitem(self, index, value)
  PyObject *self, *value;
  int index;
  { int result = -1;
    struct buffer_seg *seg;
    char *s;
    int i;

    if (index < 0)
      index += Pbuffer_len(self);
    if (value == NULL)
      result = Pbuffer__delitem(self, index);
    else if (!PyString_Check(value)) {
      PyErr_SetString(PyExc_TypeError, "excepting string of size one");
    } else if (PyString_Size(value) != 1) {
      PyErr_SetString(PyExc_TypeError, "excepting string of size one");
    } else {
      s = PyString_AsString(value);
      if (0 < index && index < Pbuffer_len(self)) {
        i = Pbuffer_position(self, index, &seg);
        if (i < 0)
          (void)AssertionError("index out of range");
        else {
          seg->p[index - i] = s[0];
          result = 0;
        }
      } else
        PyErr_SetString(PyExc_IndexError, "index out of range");
    }
    return result;
  }

static int
Pbuffer__delslice(self, low, high)
  PyObject *self;
  int low, high;
  { int result = -1;
    struct buffer_seg *seg, *loseg, *hiseg;
    int len, i, j;

    /* special case - delete all of it */
    if (low <= 0 && high >= Pbuffer_len(self)) {
      seg = Pbuffer_seg(self);
      while (seg != NULL)
        seg = destroy_bufseg(seg);
      Pbuffer_seg(self) = make_bufseg(0);
      Pbuffer_nseg(self) = 1;
      Pbuffer_len(self) = 0;
      result = 0;
    } else {
      len = high - low;
      i = Pbuffer_position(self, low, &seg);
      if (i < 0)
        (void)AssertionError("index out of range");
      else if ((loseg = break_bufseg_at(seg, i)) != NULL) {
        Pbuffer_nseg(self)++;
        /* delete the rest */
        if (high >= Pbuffer_len(self)) {
          while (loseg != NULL) {
            loseg = destroy_bufseg(loseg);
            Pbuffer_nseg(self)--;
          }
          seg->next = NULL;
          Pbuffer_len(self) -= len;
          result = 0;

        /* delete in the middle */
        } else {
          j = i = low;
          hiseg = loseg;
          j = Pbuffer_position(self, high, &hiseg);
          if (j < 0)
            (void)AssertionError("index out of range");
          else if ((seg->next = break_bufseg_at(hiseg, j)) != NULL) {
            Pbuffer_nseg(self)++;
            while (loseg != NULL && loseg != seg->next) {
              loseg = destroy_bufseg(loseg);
              Pbuffer_nseg(self)--;
            }
            Pbuffer_len(self) -= len;
            result = 0;
          }
        }
      }
    }
    return result;
  }

static int
Pbuffer__setslice(self, low, high, value)
  PyObject *self, *value;
  int low, high;
  { int result = -1;
    struct buffer_seg *seg, *newseg, *loseg, *hiseg, *tseg;
    char *s;
    int vallen, len, i, j;

    if (low < 0)
      low += Pbuffer_len(self);
    if (high < 0)
      high += Pbuffer_len(self);
    if (low > high)
      high = low = 0;
    else if (low < 0)
      low = 0;
    else if (high >= Pbuffer_len(self))
      high = Pbuffer_len(self);
    
    if (value == NULL)
      result = Pbuffer__delslice(self, low, high);
    else if (!PyString_Check(value)) {
      PyErr_SetString(PyExc_TypeError, "expecting string");
    } else {
      vallen = PyString_Size(value);
      s = PyString_AsString(value);

      len = high - low;
      i = Pbuffer_position(self, low, &seg);
      if (i < 0)
        (void)AssertionError("index out of range");
      else if ((loseg = break_bufseg_at(seg, i)) != NULL) {
        Pbuffer_nseg(self)++;
        hiseg = loseg;
        j = Pbuffer_position(self, high, &hiseg);
        if (j < 0)
          (void)AssertionError("index out of range");
        else if ((tseg = break_bufseg_at(hiseg, j)) != NULL) {
          Pbuffer_nseg(self)++;
          newseg = make_bufseg(vallen);
          write_to_bufseg(newseg, vallen, s);
          newseg->next = tseg;
          seg->next = newseg;
          Pbuffer_nseg(self) -= destroy_bufseg_chain(loseg, tseg);
          Pbuffer_len(self) += vallen - len;
          result = 0;
        }
      }
    }
    return result;
  }

PySequenceMethods Pbuffer_as_sequence = {
  Pbuffer__length,
  Pbuffer__concat,
  Pbuffer__repeat,
  Pbuffer__getitem,
  Pbuffer__getslice,
  Pbuffer__setitem,
  Pbuffer__setslice
};

static PyObject *
append_Pbuffer(self, args)
  PyObject *self, *args;
  { PyObject *result = NULL;
    struct buffer_seg *seg, *new;
    char *s;
    int sl;

    if (PyArg_ParseTuple(args, "s#", &s, &sl)) {
      if (sl > 0) {
        new = make_bufseg(sl);
        write_to_bufseg(new, sl, s);
        if (Pbuffer_seg(self) == NULL)
          Pbuffer_seg(self) = new;
        else {
          seg = Pbuffer_seg(self);
          while (seg->next != NULL)
            seg = seg->next;
          seg->next = new;
        }
        Pbuffer_nseg(self)++;
        Pbuffer_len(self) += sl;
        result = Py_None;
        Py_INCREF(result);
      } else {
        result = Py_None;
        Py_INCREF(result);
      }
    }
    return result;
  }

static PyObject *
index_Pbuffer(self, args)
  PyObject *self, *args;
  { PyObject *c;
    struct buffer_seg *seg, *curseg;
    int si, ci, found, i, len;
    char *s;

    if (!PyArg_ParseTuple(args, "S", &c))
      return NULL;
    len = PyString_Size(c);
    if (len >= Pbuffer_len(self)) {
      PyErr_SetString(PyExc_IndexError, "not found");
      return NULL;
    }
#if PYTHON_API_VERSION >= 1007
    s = PyString_AS_STRING(c);
#else
    s = PyString_AsString(c);
#endif
    seg = Pbuffer_seg(self);
    i = si = 0;
    while (seg != NULL) {
      if (si >= seg->len) {
        seg = seg->next;
        si = 0;
      }
      /* the first character matches, what about the rest? */
      if (seg->p[si] == s[0]) {
        i = 1;
        ci = si + 1;
        curseg = seg;
        found = 1;
        while (curseg != NULL && i < len) {
          if (ci >= curseg->len) {
            curseg = seg->next;
            ci = 0;
          }
          if (curseg->p[ci] != s[i]) {
            found = 0;
            break;
          }
          ci++;
          i++;
        }
        if (found)
          break;
      }
      si++;
    }
    if (!found) {
      PyErr_SetString(PyExc_IndexError, "not found");
      return NULL;
    } else
      return Py_BuildValue("i", si);
  }

/* this is to convert the buffer to one segment (Python utilities cannot
   handle more than one segment as of 1.5.2).
 */
static PyObject *
joinseg_Pbuffer(self, args)
  PyObject *self, *args;
  { struct buffer_seg *new;

    if (args != NULL && !PyArg_ParseTuple(args, ""))
      return NULL;
    if (Pbuffer_nseg(self) != 1) {
      new = join_bufseg(Pbuffer_seg(self), Pbuffer_len(self));
      Pbuffer_seg(self) = new;
      Pbuffer_nseg(self) = 1;
    }
    Py_INCREF(Py_None);
    return Py_None;
  }

/* shift data from the given position to the front of the buffer */
static PyObject *
shiftleft_Pbuffer(self, args)
  PyObject *self, *args;
  { PyObject *result = NULL;
    PyObject *p, *n;
    struct buffer_seg *seg, *newseg;
    int pos, i;

    if (!PyArg_ParseTuple(args, "i", &pos))
      return NULL;
    if (pos < 0 || pos >= Pbuffer_len(self)) {
      PyErr_SetString(PyExc_IndexError, "index out of range");
      return NULL;
    }
    i = Pbuffer_position(self, pos, &seg);
    if (i < 0)
      result = AssertionError("index out of range");
    else {
      newseg = break_bufseg_at(seg, i);
      /* start over from the beginning */
      Pbuffer_nseg(self) -=
        destroy_bufseg_chain(Pbuffer_seg(self), newseg);
      Pbuffer_seg(self) = newseg;
      Pbuffer_len(self) -= pos;
      result = Py_None;
      Py_INCREF(Py_None);
    }
    return result;
  }

PyMethodDef Pbuffer_methods[] = {
  { "append", append_Pbuffer, METH_VARARGS },
  { "index", index_Pbuffer, METH_VARARGS },
  { "join_segments", joinseg_Pbuffer, METH_VARARGS },
  { "shiftleft", shiftleft_Pbuffer, METH_VARARGS },
  { NULL, NULL },
};

/* buffer methods */

/* join the buffer segments into one, then return a pointer to the
   memory location and size */
static int
Pbuffer__getcharbuf(self, idx, pp)
  PyObject *self;
  int idx;
  void **pp;
  { PyObject *ret;

    if ((ret = joinseg_Pbuffer(self, NULL)) == NULL)
      return -1;
    Py_DECREF(ret);

    if (idx == 0) {
      *pp = Pbuffer_seg(self)->p;
      return Pbuffer_len(self);
    } else {
      PyErr_SetString(PyExc_SystemError,
                      "accessing non-existant buffer segment");
      return -1;
    }
  }

static int
Pbuffer__compare(self, other)
  PyObject *self, *other;
  { int rc = 0;
    int si, oi;
    struct buffer_seg *ss, *os;

    if (Pbuffer_len(self) == 0 && Pbuffer_len(other) == 0)
      rc = 0;
    else if (Pbuffer_len(self) == 0)
      rc = -1;
    else if (Pbuffer_len(other) == 0)
      rc = 1;
    else {
      si = oi = 0;
      ss = Pbuffer_seg(self);
      os = Pbuffer_seg(other);
      while (1) {
        /* if at end, get next segment */
        if (si >= ss->len) {
          si = 0;
          ss = (ss->next != NULL) ? ss->next : NULL;
        }
        if (oi >= os->len) {
          oi = 0;
          os = (os->next != NULL) ? os->next : NULL;
        }
        /* if not more segments, then determine who is greater */
        if (ss == NULL && os == NULL) {
          rc = 0;
          break;
        } else if (ss == NULL) {
          rc = -1;
          break;
        } else if (os == NULL) {
          rc = 1;
          break;
        }
        /* otherwise lexical comparison */
        if (ss->p[si] != os->p[oi]) {
          rc = (ss->p[si] < os->p[oi]) ? -1 : 1;
          break;
        }
        si++;
        oi++;
      }
    }
    return rc;
  }

static PyObject *
Pbuffer__getattr(self, attrname)
  PyObject *self;
  char *attrname;
  { PyObject *result = NULL;

    result = Py_FindMethod(Pbuffer_methods, self, attrname);
    return result;
  }

static int
Pbuffer__setattr(self, attrname, value)
  PyObject *self, *value;
  char *attrname;
  { int result = -1;

    return result;
  }

static PyObject *
Pbuffer__repr(self)
  PyObject *self;
  { char buf[100], str[110];
    struct buffer_seg *seg;
    int count, i, j;

    seg = Pbuffer_seg(self);
    count = i = j = 0;
    while (seg != NULL && count < Pbuffer_len(self) && count < 25) {
      if (i >= seg->len) {
        seg = seg->next;
        i = 0;
      }
      if (seg->p[i] == '\000')
        break;
      if (' ' <= seg->p[i] && seg->p[i] <= '~')
        str[j++] = seg->p[i++];
      else {
        str[j++] = '\\';
        str[j++] = '0' + ((seg->p[i] >> 6) & 07);
        str[j++] = '0' + ((seg->p[i] >> 3) & 07);
        str[j++] = '0' + ( seg->p[i]       & 07);
        i++;
      }
      count++;
    }
    if (count >= 25) {
      count = 3;
      while (count-- > 0)
        str[j++] = '.';
    }
    str[j] = '\000';
    sprintf(buf, "<Pbuffer len=%d \"%s\">", Pbuffer_len(self), str);
    return PyString_FromString(buf);
  }

static PyObject *
Pbuffer__str(self)
  PyObject *self;
  { PyObject *result = NULL;
    char *s;
    struct buffer_seg *seg;
    int i, pos;

    s = ExpPy_calloc(char, Pbuffer_len(self)+1);
    if (s == NULL)
      PyErr_NoMemory();
    else {
      seg = Pbuffer_seg(self);
      i = pos = 0;
      while (seg != NULL && i < Pbuffer_len(self)) {
        if (pos >= seg->len) {
          seg = seg->next;
          pos = 0;
        }
        s[i++] = seg->p[pos++];
      }
      s[i] = '\000';
      result = PyString_FromStringAndSize(s, i);
      ExpPy_free(s);
    }
    return result;
  }

static PyTypeObject Pbuffer_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "Pbuffer",
  sizeof(Pbuffer),
  0,
  Pbuffer__dealloc,       /* tp_dealloc */
  0,                      /* tp_print */
  Pbuffer__getattr,       /* tp_getattr */
  Pbuffer__setattr,       /* tp_setattr */
  Pbuffer__compare,       /* tp_compare */
  Pbuffer__repr,          /* tp_repr */
  0,                      /* tp_as_number */
  &Pbuffer_as_sequence,   /* tp_as_sequence */
  0,                      /* tp_as_mapping */
  0,                      /* tp_hash */
  0,                      /* tp_call */
  Pbuffer__str,           /* tp_str */
  0,                      /* tp_getattro */
  0,                      /* tp_setattro */
#if PYTHON_API_VERSION >= 1007
  0,                      /* tp_as_buffer */
  Py_TPFLAGS_DEFAULT,     /* tp_flags */
#else
  0,                      /* tp_unused */
  0,                      /* tp_xxx4 */
#endif
  0                       /* tp_doc */
};

PyObject *
new_Pbuffer(self, args)
  PyObject *self, *args;
  { char *str;
    int len;

    if (PyArg_ParseTuple(args, "s#", &str, &len))
      ;
    else if (PyArg_ParseTuple(args, "i", &len))
      str = NULL;
    else {
      PyErr_SetString(PyExc_TypeError, "expecting string or integer");
      return NULL;
    }
    return Pbuffer_NEW(str, len);
  }

#define acceptable(item) (PyString_Check(item) || \
                          strcmp((item)->ob_type->tp_name, "Tclregexp") || \
                          (PyInstance_Check(item) && \
                           PyObject_HasAttrString(item, "match")))

static PyObject *
construct_mapping(mapping)
  PyObject *mapping;
  { PyObject *map, *newitem;
    PyObject *item, *s, *c, *keylist;
    int i, l;

    if (acceptable(mapping)) {
      map = Py_BuildValue("((OO))", mapping, Py_None);

    } else if (PyTuple_Check(mapping) || PyList_Check(mapping)) {
      l = PySequence_Length(mapping);
      map = PyTuple_New(l);

      for (i=0; i < l; i++) {
        item = PySequence_GetItem(mapping, i);
        if (acceptable(item)) {
          newitem = Py_BuildValue("(OO)", item, Py_None);
          PyTuple_SetItem(map, i, newitem);
        } else if (PyTuple_Check(item) && PyTuple_Size(item) == 2) {
          s = PyTuple_GetItem(item, 0);
          c = PyTuple_GetItem(item, 1);
          if (!acceptable(s)) {
            PyErr_SetString(PyExc_TypeError, "Expecting string or regexp");
            Py_DECREF(map);
            return NULL;
          }
          newitem = Py_BuildValue("(OO)", s, c);
          PyTuple_SetItem(map, i, newitem);
        } else {
          Py_DECREF(map);
          PyErr_SetString(PyExc_TypeError, "Expecting two element tuple");
          return NULL;
        }
      }

    } else if (PyDict_Check(mapping)) {
      l = PyDict_Size(mapping);
      map = PyTuple_New(l);
      i = 0;
      while (PyDict_Next(mapping, &i, &s, &c)) {
        if (!acceptable(s)) {
          PyErr_SetString(PyExc_TypeError, "Expecting string or regexp");
          Py_DECREF(map);
          return NULL;
        }
        newitem = Py_BuildValue("(OO)", s, c);
        PyTuple_SetItem(map, i, newitem);
      }

    } else if (PyObject_HasAttrString(mapping, "keys")) {
      keylist = PyObject_CallMethod(mapping, "keys", NULL);
      l = PyList_Size(keylist);
      map = PyTuple_New(l);
      i = 0;
      while (i < l) {
        s = PyList_GetItem(keylist, i);
        if (!acceptable(s)) {
          PyErr_SetString(PyExc_TypeError, "Expecting string");
          Py_DECREF(map);
          map = NULL;
          break;
        }
        c = PyObject_GetItem(mapping, s);
        newitem = Py_BuildValue("(OO)", s, c);
        PyTuple_SetItem(map, i, newitem);
      }
      Py_DECREF(keylist);

    /* should find a _conventient_ method of handling generic sequences */
    } else {
      PyErr_SetString(PyExc_TypeError, "do not know how to convert");
      map = NULL;
    }
    return map;
  }

#undef acceptable

static PyObject *
match_string(map, buffer, reallen)
  PyObject *map;
  PyObject *buffer;
  int reallen;
  { PyObject *result;
    char *string;
    int maplen, i, pos, fpos, buflen;
    PyObject *s, *strmod, *rc;
    PyObject *itemtuple, *patt, *callback;
    static PyObject *strind;

    if (strind == NULL) {
      if ((strmod = PyImport_ImportModule("string")) == NULL)
        return NULL;
      strind = PyObject_GetAttrString(strmod, "index");
      Py_DECREF(strind);
    }
    if (strind == NULL)
      return NULL;

    if ((buflen = Pbuffer__getcharbuf(buffer, 0, &string)) == -1)
      return NULL;
    if (!PyTuple_Check(map)) {
      return AssertionError("not given a tuple");
    }

    result = NULL;
    maplen = PyTuple_Size(map);
    for (pos = 0; pos < reallen && result == NULL; pos++) {
      s = PyString_FromStringAndSize(&string[pos], reallen - pos);
      for (i = 0; i < maplen && result == NULL; i++) {
        itemtuple = PyTuple_GetItem(map, i);
        PyArg_ParseTuple(itemtuple, "OO", &patt, &callback);

        if (PyString_Check(patt)) {
          Py_INCREF(patt);
          rc = PyObject_CallFunction(strind, "SO", s, patt);
          Py_DECREF(patt);
          if (rc != NULL && PyInt_Check(rc) &&
#if PYTHON_API_VERSION >= 1007
              PyInt_AS_LONG(rc) == 0)
#else
              PyInt_AsLong(rc) == 0)
#endif
            result = Py_BuildValue("iiO", pos, i, Py_None);
          Py_XDECREF(rc);
        } else {
          Py_INCREF(patt);
          rc = PyObject_CallMethod(patt, "match", "S", s);
          Py_DECREF(patt);
          if (rc != NULL)
            result = Py_BuildValue("iiO", pos, i, rc);
          Py_XDECREF(rc);
        }
      }
      Py_DECREF(s);
    }
    if (result == NULL)
      PyErr_SetString(PyExc_IndexError, "not matched");
    return result;
  }

/* this is now a Python object since it might be used in expanded
   functionality of the interact function
 */
typedef struct stream {
  PyObject_HEAD
  int rd_strm;
  int wr_strm;
  PyObject *buffer;
  int length;
  unsigned char
        full :1,
        eof  :1,
        error:1,
        unused:5;
} StreamObject;

staticforward PyTypeObject StrmObj_Type;
#define is_StrmObj(v) ((v)->ob_type == &StrmObj_Type)
#define Stream_reader(v)  (((StreamObject *)(v))->rd_strm)
#define Stream_writer(v)  (((StreamObject *)(v))->wr_strm)
#define Stream_length(v)  (((StreamObject *)(v))->length)
#define Stream_buffer(v)  (((StreamObject *)(v))->buffer)
#define Stream_full(v)    (((StreamObject *)(v))->full)
#define Stream_eof(v)     (((StreamObject *)(v))->eof)
#define Stream_error(v)   (((StreamObject *)(v))->error)

static char StrmType__doc__[] =
  "A stream object for interact function";

/* shorthand for sockets where I/O are the same fd */
#define make_stream1(fd) make_stream2(fd, fd)

static PyObject *
make_stream2(infd, outfd)
  int infd, outfd;
  { StreamObject *s;

    s = PyObject_NEW(StreamObject, &StrmObj_Type);
    if (s != NULL) {
      Stream_reader(s) = infd;
      Stream_writer(s) = outfd;
      Stream_length(s) = 0;
      Stream_full(s) = Stream_eof(s) = Stream_error(s) = 0;
      Stream_buffer(s) = Pbuffer_NEW(NULL, STRM_BUFLEN);
    }
    return (PyObject *)s;
  }

static void
Strm_dealloc(self)
  PyObject *self;
  {
    Py_XDECREF(Stream_buffer(self));
    Stream_buffer(self) = NULL;
    PyMem_DEL(self);
  }

static char *
strm_repr(s)
  PyObject *s;
  { static char buffer[256];
    sprintf(buffer, "<STRM %d:%d (full: %d; eof: %d; error: %d) %d>",
      Stream_reader(s),
      Stream_writer(s),
      Stream_full(s),
      Stream_eof(s),
      Stream_error(s),
      Stream_length(s));
    return buffer;
  }

static int
Strm_print(self, f, dummy)
  PyObject *self;
  FILE *f;
  int dummy;
  {
    fprintf(f, strm_repr(self));
    return 0;
  }

static PyObject *
Strm_repr(self)
  PyObject *self;
  {
    return PyString_FromString(strm_repr(self));
  }

/* strm_read_buffer
   read from the reader, but store in writer's buffer
 */
static int
strm_read_buffer(reader, writer, map, callback)
  PyObject *reader, *writer;
  PyObject *map;
  PyObject **callback;
  { int length = 0;
    char *s;

    Stream_error(reader) = Stream_error(writer) = 0;
    if (is_null(Stream_reader(reader)))  /* should not be reached */
      ;
    else if (!Stream_full(writer)) {
      /* get a new buffer */
      Py_XDECREF(Stream_buffer(writer));
      Stream_buffer(writer) = Pbuffer_NEW(NULL, STRM_BUFLEN);

      length = Pbuffer__getcharbuf(Stream_buffer(writer), 0, &s);
      if (length == -1) {
        Stream_error(writer) = 1;
        return -1;
      }

      Py_BEGIN_ALLOW_THREADS
      length = read(Stream_reader(reader), s, length);
      Py_END_ALLOW_THREADS

      if (length > 0) {
        /* writes directly into the buffer */
        Stream_length(writer) = length;
        Stream_full(writer) = 1;
        /* truncate the rest of the buffer */
        PySequence_DelSlice(Stream_buffer(writer), length, STRM_BUFLEN);

      /* on Linux and maybe some other systems, terminals give EIO on EOF */
      } else if (length == 0 || errno == EIO)
        Stream_eof(reader) = 1;
      else {
        Stream_error(reader) = 1;
      }
    }
    return length;
  }

/* strm_write_buffer
   write buffer to stream, when there is data to write
   if only a partial buffer was written, then readjust
 */
static int
strm_write_buffer(reader, writer)
  struct stream *reader, *writer;
  { PyObject *dummy;
    int i, length = 0;
    char *s = NULL;

    Stream_error(writer) = 0;
    if (is_null(Stream_writer(writer)))
      Stream_full(writer) = 0;
    else if (Stream_full(writer)) {
      length = Pbuffer__getcharbuf(Stream_buffer(writer), 0, &s);
      if (length == -1) {
        Stream_error(writer) = 1;
        return -1;
      }

      Py_BEGIN_ALLOW_THREADS
      length = write(Stream_writer(writer), s, length);
      Py_END_ALLOW_THREADS

      if (length == Stream_length(writer)) {
        Stream_length(writer) = Stream_full(writer) = 0;
      } else if (length >= 0) {
        /* move unwritten data to beginning of buffer */
        dummy = PyObject_CallMethod(Stream_buffer(writer),
                                    "shiftleft", "i", length);
        Py_XDECREF(dummy);
        Stream_length(writer) -= length;
      } else
        Stream_error(writer) = 1;
    }
    return length;
  }

#ifdef STREAM_HAS_READ_AND_WRITE
static PyObject *
Strm_read(self, args)
  StreamObject *self;
  PyObject *args;
  { StreamObject *writer;
    int length;

    if (!PyArg_ParseTuple(arg, "O", &writer))
      return NULL;
    if (!is_StrmObj(writer)) {
      PyErr_SetString(PyExc_TypeError, "expecting stream object");
      return NULL;
    }
    length = strm_read_buffer(self, writer, NULL, NULL);
    return PyInt_FromLong(length);
  }

static PyObject *
Strm_write(self, args)
  StreamObject *self;
  PyObject *args;
  { StreamObject *reader;
    int length;

    if (!PyArg_ParseTuple(args, "O", &reader))
      return NULL;
    if (!is_StrmObj(reader) && reader != Py_None) {
      PyErr_SetString(PyExc_TypeError, "expecting stream object");
      return NULL;
    }
    length = strm_write_buffer(reader, self);
    return PyInt_FromLong(length);
  }

/* the stream object's methods */
static struct PyMethodDef StrmObj_methods[] = {
  { "read", (PyCFunction)Strm_read, METH_VARARGS },
  { "write", (PyCFunction)Strm_write, METH_VARARGS },
  { NULL, NULL },
};
#endif STREAM_HAS_READ_AND_WRITE

static PyObject *
Strm_getattr(self, name)
  PyObject *self;
  char *name;
  { PyObject *result;

    if (strcmp(name, "reader") == 0)
      if (is_null(Stream_reader(self))) {
        result = Py_None;
        Py_INCREF(result);
      } else
        result = PyInt_FromLong(Stream_reader(self));
    else if (strcmp(name, "writer") == 0)
      if (is_null(Stream_writer(self))) {
        result = Py_None;
        Py_INCREF(result);
      } else
        result = PyInt_FromLong(Stream_writer(self));
    else if (strcmp(name, "eof") == 0)
      result = PyInt_FromLong(Stream_eof(self));
    else if (strcmp(name, "error") == 0)
      result = PyInt_FromLong(Stream_error(self));
    else if (strcmp(name, "buffer") == 0)
      if (Stream_buffer(self) == NULL) {
        result = Py_None;
        Py_INCREF(result);
      } else {
        result = Stream_buffer(self);
      }
    else if (strcmp(name, "__members__") == 0)
      result = Py_BuildValue("[sssss]",
        "buffer", "eof", "error", "reader", "writer");
    else {
#ifdef STREAM_HAS_READ_AND_WRITE
      result = Py_FindMethod(StrmObj_methods, (PyObject *)self, name);
#else
      PyErr_SetString(PyExc_AttributeError, name);
      result = NULL;
#endif
    }
    return result;
  }

static PyTypeObject StrmObj_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,                              /*ob_size*/
  "stream",                       /*tp_name*/
  sizeof(StreamObject),           /*tp_basicsize*/
  0,                              /*tp_itemsize*/
  /* methods */
  (destructor)Strm_dealloc,       /*tp_dealloc*/
  (printfunc)Strm_print,          /*tp_print*/
  (getattrfunc)Strm_getattr,      /*tp_getattr*/
  (setattrfunc)0,                 /*tp_setattr*/
  (cmpfunc)0,                     /*tp_compare*/
  (reprfunc)Strm_repr,            /*tp_repr*/
  0,                              /*tp_as_number*/
  0,                              /*tp_as_sequence*/
  0,                              /*tp_as_mapping*/
  (hashfunc)0,                    /*tp_hash*/
  (ternaryfunc)0,                 /*tp_call*/
  (reprfunc)0,                    /*tp_str*/
  /* Space for future expansion */
  0L, 0L, 0L, 0L,
  StrmType__doc__                 /* Documentation string */
};

PyObject *
new_Strm(self, args)
  PyObject *self, *args;
  { PyObject *newobj;
    PyObject *instrm, *outstrm;
    int infd, outfd;

    instrm = outstrm = NULL;
    if (!PyArg_ParseTuple(args, "O|O", &instrm, &outstrm))
      return NULL;
    infd = get_fileno(instrm);
    if (outstrm != NULL) {
      outfd = get_fileno(outstrm);
      newobj = make_stream2(infd, outfd);
    } else
      newobj = make_stream1(infd);
    return newobj;
  }

static PyObject *
handle_callback(stream, map, callinfo)
  PyObject *stream, *map, *callinfo;
  { PyObject *obj, *t, *patt, *callback, *buffer;
    PyObject *result, *tmp, *replace;
    int pos, i, l, start, end;

    buffer = Stream_buffer(stream);
    if (!PyArg_Parse(callinfo, "(iiO)", &pos, &i, &obj)) {
      PyErr_SetString(PyExc_SystemError, "invalid return value");
      return NULL;
    }
    t = PyTuple_GetItem(map, i);
    PyArg_Parse(t, "(OO)", &patt, &callback);
    /* string */
    if (PyString_Check(patt)) {
      l = PyString_Size(patt);

    /* Tcl regexp */
    } else if (strcmp(patt->ob_type->tp_name, "Tclregexp")) {
      if (PyInt_Check(obj) && PyObject_IsTrue(obj)) {
        tmp = PyObject_CallMethod(patt, "group", "i", 0);
        PyArg_Parse(tmp, "(ii)", &start, &end);
        Py_DECREF(tmp);
        l = end - start;
      } else {
        PyErr_SetString(PyExc_ValueError, "not a valid return value");
        return NULL;
      }

    /* PCRE regexp */
    } else if (PyInstance_Check(patt)) {
      if (obj == Py_None) {
        ;
      /* PCRE */
      } else if (PyInstance_Check(obj)) {
        tmp = PyObject_GetAttrString(obj, "endpos");
        l = PyInt_AsLong(tmp);
        Py_DECREF(tmp);
      /* TclRE */
      } else if (PyInt_Check(obj) && PyInt_AsLong(obj) != 0) {
        tmp = PyObject_CallMethod(patt, "group", "i", 0);
        PyArg_Parse(tmp, "(ii)", &start, &end);
        Py_DECREF(tmp);
        l = end - start;
      } else {
        PyErr_SetString(PyExc_ValueError, "not a valid return value");
        return NULL;
      }
    
    } else {
      PyErr_SetString(PyExc_TypeError, "not a valid pattern");
      return NULL;
    }

    result = PyInt_FromLong(1);
    Py_INCREF(result);
    if (callback == Py_None) {
      PySequence_DelSlice(buffer, pos, pos + l);
      Stream_length(stream) -= l;
    } else if (PyString_Check(callback)) {
      PySequence_SetSlice(buffer, pos, pos + l, callback);
      Stream_length(stream) += (PyString_Size(callback) - l);
    } else if (PyCallable_Check(callback)) {
      tmp = PyObject_CallFunction(callback, "S", patt);
      if (tmp == NULL) {
        Py_DECREF(result);
        result = NULL;
      } else if (PyString_Check(tmp)) {
        PySequence_SetSlice(buffer, pos, pos + l, tmp);
        Stream_length(stream) += (PyString_Size(tmp) - l);
      } else {
        Py_DECREF(result);
        result = tmp;
        Py_INCREF(tmp);
      }
      Py_XDECREF(tmp);
    } else {
      Py_DECREF(result);
      result = callback;
      Py_INCREF(result);
    }
    return result;
  }

static int
set_pair(rset, wset, rst, wst)
  fd_set *rset, *wset;
  struct stream *rst, *wst;
  {
    if (!Stream_eof(rst) && Stream_eof(wst))
      return 0;
    else if (is_null(Stream_reader(rst)) && Stream_eof(wst))
      return 0;
    else if (is_null(Stream_writer(wst)) && Stream_eof(rst))
      return 0;
    else if (is_null(Stream_reader(rst)) && is_null(Stream_writer(wst)))
      Stream_eof(rst) = 1;
    else if (is_null(Stream_reader(rst)) || is_null(Stream_writer(wst)))
      return 1;
    else if (Stream_full(wst)) {         /* don't read, buffer not empty */
      FD_CLR(Stream_reader(rst), rset);
      FD_SET(Stream_writer(wst), wset);
    } else if (!Stream_eof(rst)) {
      FD_SET(Stream_reader(rst), rset);
      FD_CLR(Stream_writer(wst), wset);
    } else {
      FD_CLR(Stream_reader(rst), rset);
      FD_CLR(Stream_writer(wst), wset);
    }
    return 1;
  }

/* rc = interact(input_stream, output_stream, map);
   Data read from each stream is written to the other stream.
   The return code will be
     -1    error from select(3)
     0     EOF or I/O error from read(2) or write(2)
     1     break
     2     timeout from select (based on exp_timeout value)
     3     exception from callback
 */
static int
interact(rst, wst, map)
  PyObject *rst, *wst, *map;
  { fd_set rset, wset;
    int count, rc, maxfd = 0;
    struct timeval *timeout, time;
    PyObject *result, *callback = NULL;
    PyObject *returncode;

    /* determine timeout value for select function based on Expect values */
    if (exp_timeout == -1)
      timeout = (struct timeval *)NULL;
    else {
      time.tv_sec = (time_t)exp_timeout;
      time.tv_usec = (long)0;
      timeout = &time;
    }

    /* find the largest file destriptor */
#define max(a, b) (((a) > (b)) ? (a) : (b))
    if (!is_null(Stream_reader(rst))) maxfd = max(Stream_reader(rst), maxfd);
    if (!is_null(Stream_writer(rst))) maxfd = max(Stream_writer(rst), maxfd);
    if (!is_null(Stream_reader(wst))) maxfd = max(Stream_reader(wst), maxfd);
    if (!is_null(Stream_writer(wst))) maxfd = max(Stream_writer(wst), maxfd);
#undef max
    maxfd++; /* select wants one more than max */

    FD_ZERO(&rset);
    FD_ZERO(&wset);
    while (1) {
      /* only select on streams when the buffer is not empty */
      if (!set_pair(&rset, &wset, wst, rst) ||
          !set_pair(&rset, &wset, rst, wst))
        Stream_eof(rst) = Stream_eof(wst) = 1;

      /* break on error, let caller handle it */
      if (Stream_error(rst) || Stream_error(wst))
        break;
      else if (Stream_eof(rst) && Stream_eof(wst))
        break;

      Py_BEGIN_ALLOW_THREADS
      rc = select(maxfd, &rset, &wset, NULL, timeout);
      Py_END_ALLOW_THREADS
      if (rc == -1)
        return -1;

      count = 0;
      if (!is_null(Stream_writer(rst)) &&
          FD_ISSET(Stream_writer(rst), &wset))
        count++, strm_write_buffer(wst, rst);
      if (!is_null(Stream_writer(wst)) &&
          FD_ISSET(Stream_writer(wst), &wset))
        count++, strm_write_buffer(rst, wst);
      if (!is_null(Stream_reader(rst)) &&
          FD_ISSET(Stream_reader(rst), &rset)) {
        callback = NULL;
        count++;
        if (strm_read_buffer(rst, wst, map, &callback) == -1)
          return 3; /* error */
        if (callback != NULL) {
          returncode = handle_callback(wst, map, callback);
          if (returncode == NULL)
            return 3; /* exception */
          else if (!PyObject_IsTrue(returncode)) {
            Py_DECREF(returncode);
            return 1; /* break */
          }
          Py_DECREF(returncode);
          Py_DECREF(callback);
        }
      }
      if (!is_null(Stream_reader(wst)) &&
          FD_ISSET(Stream_reader(wst), &rset)) {
        callback = NULL;
        count++;
        /* if (strm_read_buffer(wst, rst, map, &callback) == -1) */
        /* the receiver spawn does not get processed */
        if (strm_read_buffer(wst, rst, NULL, NULL) == -1)
          return 3; /* error */
        if (callback != NULL) {
          returncode = handle_callback(rst, map, callback);
          if (returncode == NULL)
            return 3; /* exception */
          else if (!PyObject_IsTrue(returncode)) {
            Py_DECREF(returncode);
            return 1; /* break */
          }
          Py_DECREF(returncode);
          Py_DECREF(callback);
        }
      }
      if (count == 0)
        return 2;  /* timeout */

      /* return an error before we try a callback */
      if (Stream_error(rst) || Stream_error(wst))
        return 0;
      callback = NULL;
    }
    return 0;
  }

static PyObject *
convert_pyarg(argument)
  PyObject *argument;
  { PyObject *result = NULL;
    PyObject *tmp, *meth;
    PyObject *t1, *t2;
    int fd1, fd2;

    if (PyTuple_Check(argument)) {
      if (!PyArg_ParseTuple(argument, "OO", &t1, &t2))
        result = NULL;
      else if (((fd1 = get_fileno(t1)) >= 0 || fd1 == NO_STREAM_NUM) &&
               ((fd2 = get_fileno(t2)) >= 0 || fd2 == NO_STREAM_NUM))
        result = make_stream2(fd1, fd2);
    } else if ((fd1 = get_fileno(argument)) >= 0 || fd1 == NO_STREAM_NUM)
      result = make_stream1(fd1);
    return result;
  }

char pyexp_interact__doc__[] = 
  "Interact function -\n\
Given two arguments, link the I/O streams between them.\n\
An I/O stream may be one of the following:\n\
* None                   - data is ignored\n\
* socket-like object     - a readable and writable object\n\
* file descriptor        - a readable and writable POSIX file descriptor\n\
* tuple (reader, writer) -\n\
  either could be None, a file-like object or a file descriptor.\n\
If None, reads are ignored and writes are not completed (with no error).";
PyObject *
pyexp_interact(self, args)
  PyObject *self, *args;
  { PyObject *instream, *outstream;
    PyObject *arg1, *arg2, *meth;
    int rc;
    extern PyObject *Expect_TimeoutError;
    PyObject *map, *mapping, *result = NULL;

    mapping = NULL;
    if (!PyArg_ParseTuple(args, "OO|O", &arg1, &arg2, &mapping))
      return NULL;

    if ((instream = convert_pyarg(arg1)) == NULL)
      return NULL;
    if ((outstream = convert_pyarg(arg2)) == NULL) {
      Py_DECREF(instream);
      return NULL;
    }
    if (mapping != NULL) {
      if ((map = construct_mapping(mapping)) == NULL)
        return NULL;
    } else
      map = NULL;

    /* only return on an exception */
    rc = interact(instream, outstream, map);

    if (rc < 0) /* select error... should we get select.error? */
      PyErr_SetFromErrno(PyExc_IOError);
    else if (rc == 1) {
      result = Py_None;
      Py_INCREF(result);
    } else if (rc == 2)
      PyErr_SetNone(Expect_TimeoutError);
    else if (rc == 3 && PyErr_Occurred())
      /* pass thru */;
    else if (rc == 3) {
      result = Py_None;
      Py_INCREF(result);
    } else if (Stream_eof(instream) && Stream_eof(outstream))
      PyErr_SetNone(PyExc_EOFError);
    else if (Stream_error(instream) || Stream_error(outstream))
      /* should be set from I/O functions */
      PyErr_SetFromErrno(PyExc_IOError);
    Py_DECREF(instream);
    Py_DECREF(outstream);
    return result;
  }

