/*  stty.c  - Termios object for a file or file descriptor  */
/* Copyright (C) 1999 Michael P. Reilly, All rights reserved
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Log: stty.c,v $
 * Revision 1.4  2000/05/04 11:33:34  arcege
 * Make the baud values more extensible.  Add some more support functions.
 * Make the stty_termios function compile-optional.
 *
 * Revision 1.3  2000/02/23 12:12:44  arcege
 * A lot of changes:
 * * Added a termios() function
 * * A redesign of access to the termios settings for greater expansion
 * * cc dict returns both cooked and raw values
 *
 */

#include "ExpPycnf.h"
#include "patchlevel.h"
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>

#include <Python.h>
#if 0  /* needed for PyObject_Print() */
#include <abstract.h>
#endif 
#include "ExpPyMem.h"

#define sizeofvnames(v) (sizeof(v)/sizeof(v[0]))

static struct baud_value_t {
  int bval;
  char *name;
  int rate;
} c_baud_values[] = {
  { B0,      "B0",          0 },
#ifdef B50
  { B50,     "B50",         50 },
#endif
#ifdef B75
  { B75,     "B75",         75 },
#endif
#ifdef B110
  { B110,    "B110",        110 },
#endif
#ifdef B134
  { B134,    "B134",        134 },
#endif
#ifdef B150
  { B150,    "B150",        150 },
#endif
#ifdef B200
  { B200,    "B200",        200 },
#endif
#ifdef B300
  { B300,    "B300",        300 },
#endif
#ifdef B600
  { B600,    "B600",        600 },
#endif
#ifdef B1200
  { B1200,   "B1200",       1200 },
#endif
#ifdef B1800
  { B1800,   "B1800",       1800 },
#endif
#ifdef B2400
  { B2400,   "B2400",       2400 },
#endif
#ifdef B4800
  { B4800,   "B4800",       4800 },
#endif
#ifdef B9600
  { B9600,   "B9600",       9600 },
#endif
#ifdef B19200
  { B19200,  "B19200",      19200 },
#endif
#ifdef B38400
  { B38400,  "B38400",      38400 },
#endif
#ifdef B57600
  { B57600,  "B57600",      57600 },
#endif
#ifdef B76800
  { B76800,  "B76800",      76800 },
#endif
#ifdef B115200
  { B115200, "B115200",     115200 },
#endif
#ifdef B153600
  { B153600, "B153600",     153600 },
#endif
#ifdef B230400
  { B230400, "B230400",     230400 },
#endif
#ifdef B307200
  { B307200, "B307200",     307200 },
#endif
#ifdef B460800
  { B460800, "B460800",     460800 },
#endif
};

static struct vname_t {
  char *name;
  int pos;
} cc_raw_vnames[] = {
#ifdef VMIN
  { "VMIN", VMIN },
#endif
#ifdef VTIME
  { "VTIME", VTIME },
#endif

}, cc_cooked_vnames[] = {

#ifdef VINTR
  { "VINTR", VINTR },
#endif
#ifdef VQUIT
  { "VQUIT", VQUIT },
#endif
#ifdef VERASE
  { "VERASE", VERASE },
#endif
#ifdef VKILL
  { "VKILL", VKILL },
#endif
#ifdef VEOF
  { "VEOF", VEOF },
#endif
#ifdef VEOL
  { "VEOL", VEOL },
#endif
#ifdef VEOL2
  { "VEOL2", VEOL2 },
#endif
#ifdef VSWTCH
  { "VSWTCH", VSWTCH },
#endif
#ifdef VSTART
  { "VSTART", VSTART },
#endif
#ifdef VSTOP
  { "VSTOP", VSTOP },
#endif
#ifdef VSUSP
  { "VSUSP", VSUSP },
#endif
#ifdef VDSUSP
  { "VDSUSP", VDSUSP },
#endif
#ifdef VREPRINT
  { "VREPRINT", VREPRINT },
#endif
#ifdef VDISCARD
  { "VDISCARD", VDISCARD },
#endif
#ifdef VWERASE
  { "VWERASE", VWERASE },
#endif
#ifdef VLNEXT
  { "VLNEXT", VLNEXT },
#endif
};

static int
convert_py_str(value, svalue)
  PyObject *value;
  char **svalue;
  { PyObject *temp;

    if (svalue == NULL) {
      PyErr_SetString(PyExc_SystemError, "need address of integer");
      return -1;
    }
    if (value == Py_None) {
      *svalue = NULL;
    } else if (PyString_Check(value)) {
#if PYTHON_API_VERSION >= 1007
      *svalue = PyString_AS_STRING(value);
#else
      *svalue = PyString_AsString(value);
#endif
    } else if ((temp = PyObject_Str(value)) != NULL) {
#if PYTHON_API_VERSION >= 1007
      *svalue = PyString_AS_STRING(temp);
#else
      *svalue = PyString_AsString(temp);
#endif
      Py_DECREF(temp);
    } else {
      PyErr_SetString(PyExc_TypeError, "expecting string");
      return -1;
    }
    return 0;
  }

static int
convert_py_int(value, ivalue, allow_none)
  PyObject *value;
  long *ivalue;
  int allow_none;
  { PyObject *temp;

    if (ivalue == NULL) {
      PyErr_SetString(PyExc_SystemError, "need address of integer");
      return -1;
    }
    if (allow_none && value == Py_None)
      *ivalue = 0;
    else if (PyInt_Check(value))
#if PYTHON_API_VERSION >= 1007
      *ivalue = PyInt_AS_LONG(value);
#else
      *ivalue = PyInt_AsLong(value);
#endif
    else if ((temp = PyNumber_Int(value)) != NULL) {
#if PYTHON_API_VERSION >= 1007
      *ivalue = PyInt_AS_LONG(temp);
#else
      *ivalue = PyInt_AsLong(temp);
#endif
      Py_DECREF(temp);
    } else {
      PyErr_SetString(PyExc_TypeError, "expecting integer");
      return -1;
    }
    return 0;
  }

/* functionality for stty object */

staticforward PyTypeObject SttyType;

typedef struct {
  PyObject_HEAD
  int fd;
  struct termios settings;
} Stty;

#define Stty_Check(v)    ((v)->ob_type == &SttyType)
#define Stty_fd(v)       (((Stty *)(v))->fd)
#define Stty_settings(v) (((Stty *)(v))->settings)

/* save data to the terminal device */
static int
Stty_read_settings(self)
  PyObject *self;
  {
    return tcgetattr(Stty_fd(self), &Stty_settings(self));
  }
static int
Stty_write_settings(self, when)
  PyObject *self;
  int when;
  {
    return tcsetattr(Stty_fd(self), when, &Stty_settings(self));
  }

/* set/get bit value based on Python truth object */
static PyObject *
get_bit_value(flags, setting, not)
  int flags, setting, not;
  { if ((!not && (flags & setting)) || (not && !(flags & setting)))
      return PyInt_FromLong(1);
    else
      return PyInt_FromLong(0);
  }
static int
set_bit_value(flags, setting, not, value)
  int *flags, setting, not;
  PyObject *value;
  {
    if ((!not && PyObject_IsTrue(value)) ||
        (not && !PyObject_IsTrue(value)))
      *flags |= setting;
    else
      *flags &= ~setting;
    return 1;
  }

static int
translate_py_to_char(value, ptr)
  PyObject *value;
  char *ptr;
  { char *str;
    if (value == Py_None)
      *ptr = '\000';
    else if (value != NULL && PyString_Check(value)) {
#if PYTHON_API_VERSION >= 1007
      str = PyString_AS_STRING(value);
#else
      str = PyString_AsString(value);
#endif
      *ptr = str[0];
    } else if (value != NULL) {
      PyErr_SetString(PyExc_TypeError, "not a string");
      return -1;
    }
    return 0;
  }

static PyObject *
translate_char_to_py(value)
  char value;
  {
    return PyString_FromStringAndSize(&value, 1);
  }
static PyObject *
translate_int_to_py(value)
  char value;
  { long i = (long)value;
    return PyInt_FromLong(i);
  }

static int
cc_get_dict(dict, name, value)
  PyObject *dict;
  char *name, *value;
  { PyObject *result;

    /* no need to handle the reference count here */
    result = PyDict_GetItemString(dict, name);
    return translate_py_to_char(result, value);
  }

static void
cc_set_dict(dict, name, value, type)
  PyObject *dict;
  char *name, value;
  int type;
  { PyObject *result;

    if (type)
      result = translate_char_to_py(value);
    else
      result = translate_int_to_py(value);
    PyDict_SetItemString(dict, name, result);
  }

static int
termios_cc_set(settings, name, value, cooked)
  struct termios *settings;
  char *name;
  PyObject *value;
  int cooked;
  { int result = 0;
    struct vname_t *names;
    char str;
    int i, l, ch;

    if (PyString_Check(value)) {
      if (translate_py_to_char(value, &str) == -1)
        return -1;
      ch = str;
    } else if (PyInt_Check(value)) {
#if PYTHON_API_VERSION >= 1007
      ch = PyInt_AS_LONG(value);
#else
      ch = PyInt_AsLong(value);
#endif
    } else {
      PyErr_SetString(PyExc_TypeError, "expected string or integer");
      return -1;
    }

    if (cooked) {
      names = cc_raw_vnames;
      l = sizeofvnames(cc_raw_vnames);
    } else {
      names = cc_cooked_vnames;
      l = sizeofvnames(cc_cooked_vnames);
    }
    for (i = 0; i < l; i++)
      if (strcmp(names[i].name, name) == 0)
        break;

    if (i >= l) {
      result = -1;
      PyErr_SetString(PyExc_KeyError, name);
    } else
      settings->c_cc[names[i].pos] = ch;
    return result;
  }

static PyObject *
termios_cc_get(settings, name, cooked)
  struct termios *settings;
  char *name;
  int cooked;
  { PyObject *result = NULL;
    struct vname_t *names;
    int i, l;

    if (cooked) {
      names = cc_cooked_vnames;
      l = sizeofvnames(cc_cooked_vnames);
    } else {
      names = cc_raw_vnames;
      l = sizeofvnames(cc_raw_vnames);
    }
    for (i = 0; i < l; i++)
      if (strcmp(names[i].name, name) == 0)
        break;

    if (i >= l)
      PyErr_SetString(PyExc_KeyError, name);
    else if (cooked)
      result = translate_char_to_py(settings->c_cc[names[i].pos]);
    else
      result = translate_int_to_py(settings->c_cc[names[i].pos]);
    return result;
  }

/* Convert values in CC to a dictionary */
static PyObject *
termios_cc_to_py(settings)
  struct termios *settings;
  { PyObject *dict, *ch;
    register int i, l;
    register char *name;

    dict = PyDict_New();
    if (dict != NULL) {
      l = sizeofvnames(cc_raw_vnames);
      for (i = 0; i < l; i++) {
        name = cc_raw_vnames[i].name;
        ch = termios_cc_get(settings, name, 0);
        PyDict_SetItemString(dict, name, ch);
      }
      l = sizeofvnames(cc_cooked_vnames);
      for (i = 0; i < l; i++) {
        name = cc_cooked_vnames[i].name;
        ch = termios_cc_get(settings, name, 1);
        PyDict_SetItemString(dict, name, ch);
      }
    }
    return dict;
  }

/* Convert values from a dictionary to a CC structure */
/* Structures not defined on compiled system are ignored */
static PyObject *
termios_py_to_cc(settings, dict)
  struct termios *settings;
  PyObject *dict;
  { PyObject *value;
    register int i, l;
    register char *name;

    l = sizeofvnames(cc_raw_vnames);
    for (i = 0; i < l; i++) {
      name = cc_raw_vnames[i].name;
      value = PyDict_GetItemString(dict, name);
      if (value != NULL)
        termios_cc_set(settings, name, value, 0);
    }
    l = sizeofvnames(cc_cooked_vnames);
    for (i = 0; i < l; i++) {
      name = cc_cooked_vnames[i].name;
      value = PyDict_GetItemString(dict, name);
      if (value != NULL)
        termios_cc_set(settings, name, value, 1);
    }
    Py_INCREF(Py_None);
    return Py_None;
  }

static int
termios_set_speed(settings, ispeed, ospeed)
  struct termios *settings;
  int ispeed, ospeed;
  {
#ifdef HAVE_CFGETISPEED
    cfsetispeed(settings, ispeed);
    cfsetospeed(settings, ospeed);
#else

#ifndef CIBAUD
    ospeed = ispeed;
#endif
#ifdef CBAUDEXT
    if (ospeed > CBAUD) {
      ospeed -= (CBAUD - 1);
      settings->c_cflag |= CBAUDEXT;
    } else
      settings->c_cflag &= ~CBAUDEXT;
#endif
    settings->c_cflag = (settings->c_cflag & ~CBAUD) | (ospeed & CBAUD);
#ifdef CIBAUD
    if (ispeed == 0)
      ispeed = ospeed;
#ifdef CIBAUDEXT
    if ((ispeed << IBSHIFT) > CIBAUD) {
      ispeed -= ((CIBAUD >> IBSHIFT) + 1);
      settings->c_cflag |= CIBAUDEXT;
    } else
      settings->c_cflag &= ~CIBAUDEXT;
#endif
    settings->c_cflag =
      (settings->c_cflag & ~CIBAUD) |
      ((ispeed << IBSHIFT) & CIBAUD);
#endif /*CIBAUD*/
#endif /*HAVE_CFGETISPEED*/
    return 0;
  }

static int
termios_get_speed(settings, ispeed, ospeed)
  struct termios *settings;
  int *ispeed, *ospeed;
  { int t_baud;

    if (ispeed == NULL)
      return 0;
#ifdef HAVE_CFGETISPEED
    *ispeed = cfgetispeed(settings);
    if (ospeed != NULL)
      *ospeed = cfgetospeed(settings);
#else

#ifdef CBAUDEXT
    if (ospeed == NULL)
      ospeed = ispeed;
    if (settings->c_cflag & CBAUDEXT)
      *ospeed = (settings->c_cflag & CBAUD) + (CBAUD + 1);
    else
#endif
      *ospeed = (settings->c_cflag & CBAUD);
#ifdef CIBAUD
    if (ispeed != ospeed) {
#ifdef CIBAUDEXT
      if (settings->c_cflag & CIBAUDEXT)
        *ispeed = ((settings->c_cflag & CIBAUD) >> IBSHIFT)
                  + (CIBAUD >> IBSHIFT) + 1;
      else
#endif
        *ispeed = (settings->c_cflag & CIBAUD) >> IBSHIFT;
    }
#endif /*CIBAUD*/
#endif /*HAVE_CFGETISPEED*/

    return 1;
  }

/* return the termio(7) value for the appropriate given baud rate.
   Check PyErr_Occurred() instead of the return value.  The Python
   object None is equivalent to B0 (hangup).
 */
static int
termios_py_to_baud(settings, value)
  struct termios *settings;
  PyObject *value;
  { int i, l, i_value;
    int ti_value, to_value;
    PyObject *temp;

    if (convert_py_int(value, &i_value, 1) == -1)
      return -1;

    l = sizeofvnames(c_baud_values);
    for (i = 0; i < l; i++)
      if (i_value == c_baud_values[i].rate) {
        ti_value = c_baud_values[i].bval;
        break;
      }
    if (i >= l) {
      PyErr_SetString(PyExc_ValueError, "baud rate not valid");
      return -1;
    }
    to_value = ti_value; /* until I get support for both */

    (void)termios_set_speed(settings, ti_value, to_value);
    return 0;
  }

/* return the baud rate as a Python object. if the baud rate is not
   an "accepted" value, return None.  If the baud rate is B0, then
   return None (it shouldn't be B0, but...)
 */
static PyObject *
termios_baud_to_py(settings)
  struct termios *settings;
  { int baudrate;
    int i, l, ti_baud;

    (void)termios_get_speed(settings, &ti_baud, NULL);

    l = sizeofvnames(c_baud_values);
    for (i = 0; i < l; i++)
      if (ti_baud == c_baud_values[i].bval) {
        baudrate = c_baud_values[i].rate;
        break;
      }
    if (i >= l || baudrate == 0) {
      Py_INCREF(Py_None);
      return Py_None;
    } else
      return PyInt_FromLong(baudrate);
  }

static char *
termios_baud_name(t_baud)
  int t_baud;
  { char *result = NULL;
    int i, l;

    l = sizeofvnames(c_baud_values);
    for (i = 0; i < l; i++)
      if (t_baud == c_baud_values[i].bval) {
        result = c_baud_values[i].name;
        break;
      }

    return result;  /* a static string */
  }

static int
termios_baud_value(t_baud)
  char *t_baud;
  { int i, l, result = -1;

    if (t_baud == NULL) return -1;
    l = sizeofvnames(c_baud_values);
    for (i = 0; i < l; i++)
      if (strcmp(t_baud, c_baud_values[i].name) == 0) {
        result = c_baud_values[i].bval;
        break;
      }
    return result;
  }

/* Return a tuple containing the available rates for this machine */
static PyObject *
termios_baud_rates()
  { PyObject *result, *temp;
    register int i, size = 7;

    size = sizeofvnames(c_baud_values);
    result = PyTuple_New(size);
    for (i = 0; i < size; i++)
      if (c_baud_values[i].bval == B0) {
        PyTuple_SetItem(result, i, Py_None);
        Py_INCREF(Py_None);
      } else {
        temp = PyInt_FromLong(c_baud_values[i].rate);
        PyTuple_SetItem(result, i, temp);
      }
    return result;
  }

static PyObject *
get_stty_value(settings, name)
  struct termios *settings;
  char *name;
  { PyObject *result, *ccdict;
    int i_value;

    if (strcmp(name, "baud") == 0) {
      result = termios_baud_to_py(settings);
    } else if (strcmp(name, "raw") == 0) {
      result = get_bit_value(settings->c_lflag, ICANON, 1);
    } else if (strcmp(name, "cooked") == 0) {
      result = get_bit_value(settings->c_lflag, ICANON, 0);
    } else if (strcmp(name, "echo") == 0) {
      result = get_bit_value(settings->c_lflag, ECHO, 0);
    } else if (strcmp(name, "cc") == 0) {
      result = termios_cc_to_py(settings);
    } else if (name[0] == 'V') {
      ccdict = termios_cc_to_py(settings);
      result = PyDict_GetItemString(ccdict, name);
      if (result == NULL)
        PyErr_SetString(PyExc_ValueError, "unknown setting");
      Py_DECREF(ccdict);
      Py_INCREF(result);
    } else if (strcmp(name, "_cflag") == 0 ||
               strcmp(name, "_iflag") == 0 ||
               strcmp(name, "_lflag") == 0 ||
               strcmp(name, "_oflag") == 0) {
      switch (name[1]) {
        case 'c': i_value = settings->c_cflag; break;
        case 'i': i_value = settings->c_iflag; break;
        case 'l': i_value = settings->c_lflag; break;
        case 'o': i_value = settings->c_oflag; break;
      }
      result = PyInt_FromLong(i_value);
    } else {
      PyErr_SetString(PyExc_AttributeError, name);
      return NULL;
    }
    return result;
  }

static int
set_stty_value(settings, name, value)
  struct termios *settings;
  char *name;
  PyObject *value;
  { PyObject *ccdict, *f;
    int i_value;

    if (strcmp(name, "baud") == 0) {
      PyErr_Clear();
      termios_py_to_baud(settings, value);
      if (PyErr_Occurred())
        return 0;
    } else if (strcmp(name, "raw") == 0) {
      set_bit_value(&(settings->c_lflag), ICANON, 1, value);
    } else if (strcmp(name, "cooked") == 0) {
      set_bit_value(&(settings->c_lflag), ICANON, 0, value);
    } else if (strcmp(name, "echo") == 0) {
      set_bit_value(&(settings->c_lflag), ECHO, 0, value);
    } else if (strcmp(name, "_cflag") == 0 ||
               strcmp(name, "_iflag") == 0 ||
               strcmp(name, "_lflag") == 0 ||
               strcmp(name, "_oflag") == 0) {
      if (convert_py_int(value, &i_value, 0) == -1)
        return 0;
      switch (name[1]) {
        case 'c': settings->c_cflag = i_value; break;
        case 'i': settings->c_iflag = i_value; break;
        case 'l': settings->c_lflag = i_value; break;
        case 'o': settings->c_oflag = i_value; break;
      }
    } else if (name[0] == 'V') {
      ccdict = Py_BuildValue("{s: O}", name, value);
      f = termios_py_to_cc(settings, ccdict);
      Py_DECREF(ccdict);
      if (f == NULL)
        return 0;
      Py_DECREF(f);
    } else {
      PyErr_SetString(PyExc_ValueError, "unknown setting");
      return 0;
    }
    return 1;
  }

static int
Stty_parse_stty_dict(self, keywds)
  PyObject *self, *keywds;
  { int result, i, l;
    struct termios *settings;
    PyObject *key, *value;
    char *pname;

    settings = &Stty_settings(self);
    result = 1;
    l = PyDict_Size(keywds);
    for (i = 0; i < l;) {
      /* PyDict_Next() increments the index */
      if (!PyDict_Next(keywds, &i, &key, &value)) {
        PyErr_SetString(PyExc_SystemError,
                        "unknown error in keyword argument");
        result = 0;
        break;
      }
      if (!PyString_Check(key)) {
        PyErr_SetString(PyExc_TypeError, "expecting string for keyword");
        result = 0;
        return 0;
      }

#if PYTHON_API_VERSION >= 1007
      pname = PyString_AS_STRING(key);
#else
      pname = PyString_AsString(key);
#endif
      if (!set_stty_value(settings, pname, value)) {
        result = 0;
        break;
      }

    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return result;
  }

/* parse the given tuple as stty arguments and set the appropriate bits
   in the termios structure.
 */
static int
Stty_parse_stty_list(self, list)
  PyObject *self, *list;
  { int result, i, l, need_arg = 0;
    struct termios *settings;
    PyObject *dict, *obj, *temp;
    PyObject *pytrue, *pyfalse;
    char *pname;

    pyfalse= PyInt_FromLong(0);
    pytrue = PyInt_FromLong(1);
    dict = PyDict_New();
    settings = &Stty_settings(self);
    l = PyTuple_Size(list);
    for (i = 0; i < l; i++) {
#if PYTHON_API_VERSION >= 1007
      obj = PyTuple_GET_ITEM(list, i);
#else
      obj = PyTuple_GetItem(list, i);
#endif

      /* if a parameter needs an argument, set the value here */
      if (need_arg) {
        need_arg = 0;
        if (strcmp(pname, "baud") == 0) {
          PyDict_SetItemString(dict, "buad", obj);
        }

      /* get the parameter name
         if applicable, set the boolean value here
         otherwise, set to get the next argument
       */
      } else {
        if (obj == Py_None) {
          continue; /* ignore */
        } else if (convert_py_str(obj, &pname) == -1) {
          Py_DECREF(pytrue); Py_DECREF(pyfalse);
          Py_DECREF(dict);
          return 0;
        }

        if (strcmp(pname, "raw") == 0 || strcmp(pname, "-cooked") == 0)
          PyDict_SetItemString(dict, "cooked", pyfalse);
        else if (strcmp(pname, "cooked") == 0 || strcmp(pname, "-raw") == 0)
          PyDict_SetItemString(dict, "cooked", pytrue);
        else if (strcmp(pname, "echo") == 0)
          PyDict_SetItemString(dict, "echo", pytrue);
        else if (strcmp(pname, "-echo") == 0)
          PyDict_SetItemString(dict, "echo", pyfalse);
        else if (strcmp(pname, "baud") == 0)
          need_arg = 1;
      }
    }
    if (need_arg) {
      PyErr_SetString(PyExc_ValueError, "argument expected");
      return 0;
    }
    result = Stty_parse_stty_dict(self, dict);
    Py_DECREF(dict);
    return result;
  }

/* I couldn't make up my mind to return a dict or a tuple from stty() */

#ifdef STTY_RETURNS_DICT
/* return a Python dictionary populated with the values from the
   file descriptor's termio structure
 */
static PyObject *
Stty_unparse_stty_dict(self)
  PyObject *self;
  { PyObject *dict, *key, *value;
    struct termios *settings;

    if (Stty_read_settings(self) == -1)
      return PyErr_SetFromErrno(PyExc_IOError);
    settings = &Stty_settings(self);

    if ((dict = PyDict_New()) == NULL)
      return NULL;

    /* reference is "borrowed" */
    value = get_stty_value(settings, "echo");
    PyDict_SetItemString(dict, "echo", value);

    value = get_stty_value(settings, "cooked");
    PyDict_SetItemString(dict, "cooked", value);

    value = get_stty_value(settings, "baud");
    PyDict_SetItemString(dict, "baud", value);
    return dict;
  }
#else /* STTY_RETURNS_DICT */

/* return a Python tuple populated with the values from the
   file descriptor's termio structure
 */
static PyObject *
Stty_unparse_stty_tuple(self)
  PyObject *self;
  { PyObject *tuple, *value;
    struct termios *settings;

    if (Stty_read_settings(self) == -1)
      return PyErr_SetFromErrno(PyExc_IOError);

    settings = &Stty_settings(self);
    if ((tuple = PyTuple_New(4)) == NULL)
      return NULL;

    value = PyString_FromString(
      settings->c_lflag & ECHO ? "echo" : "-echo"
    );
    if (value == NULL) {
      Py_DECREF(tuple);
      return NULL;
    }
    Py_INCREF(value);
#if PYTHON_API_VERSION >= 1007
    PyTuple_SET_ITEM(tuple, 0, value);
#else
    PyTuple_SetItem(tuple, 0, value);
#endif

    value = PyString_FromString(
      settings->c_lflag & ICANON ? "cooked" : "raw"
    );
    if (value == NULL) {
      Py_DECREF(tuple);
      return NULL;
    }
    Py_INCREF(value);
#if PYTHON_API_VERSION >= 1007
    PyTuple_SET_ITEM(tuple, 1, value);
#else
    PyTuple_SetItem(tuple, 1, value);
#endif

    value = PyString_FromString("baud");
    if (value == NULL) {
      Py_DECREF(tuple);
      return NULL;
    }
    Py_INCREF(value);
    PyTuple_SetItem(tuple, 2, value);
    value = termios_baud_to_py(settings);
    if (value == NULL) {
      Py_DECREF(tuple);
      return NULL;
    }
    PyTuple_SetItem(tuple, 3, value);

    return tuple;
  }
#endif /* STTY_RETURNS_DICT */

#ifdef STTY_TERMIOS_FUNCTION

#define GET_NUM2NAM 0
#define GET_NAM2NUM 1
#define GET_BAUDNAM 2
#define GET_BAUDNUM 3

static PyObject *
stty_termios_mod_info(settings, value_type)
  struct termios *settings;
  int value_type;
  { PyObject *result = NULL;
    static PyObject *termios_mod = NULL;
    static PyObject *termios_num2nam = NULL;
    static PyObject *termios_nam2num = NULL;
    static PyObject *termios_baudnam = NULL;
    static PyObject *termios_baudnum = NULL;
    PyObject *mod_dict, *key, *value;
    int i, l, cont = 0;
    char *keyname, *digs, *digits = "0123456789";

    /* a lot of prep-work getting "global" values set up */
    if (termios_mod == NULL) {
      termios_mod = PyImport_ImportModule("TERMIOS");
    }
    if (termios_mod == NULL) {
      if (!(PyErr_Occurred() && PyErr_ExceptionMatches(PyExc_ImportError)))
        ;/*PyErr_SetString(PyExc_ImportError, "No module named TERMIOS");*/
      return NULL;
    }

    if (termios_baudnam == NULL) {
      termios_baudnam = PyDict_New();
      if (termios_baudnam == NULL)
        return NULL;
      mod_dict = PyModule_GetDict(termios_mod);
      i = 0;
      while (PyDict_Next(mod_dict, &i, &key, &value)) {
        keyname = PyString_AsString(key);
        if (keyname[0] != 'B')
          continue;
        for (digs = digits, cont = 0; *digs; digs++)
          if (keyname[1] != *digs) {
            cont = 1;
            break;
          }
        if (!cont)
          continue;
        PyDict_SetItem(termios_baudnam, key, value);
      }
    }

    if (termios_baudnum == NULL) {
      termios_baudnum = PyDict_New();
      if (termios_baudnum == NULL)
        return NULL;
      i = 0;
      while (PyDict_Next(termios_baudnam, &i, &key, &value)) {
        PyDict_SetItem(termios_baudnum, value, key);
      }
    }

    if (termios_nam2num == NULL) {
      termios_nam2num = PyDict_New();
      if (termios_nam2num == NULL) {
        return NULL;
      }
      l = sizeofvnames(cc_cooked_vnames);
      for (i = 0; i < l; i++) {
        value = PyObject_GetAttrString(termios_mod,
          cc_cooked_vnames[i].name);
        if (value != NULL) {
          Py_INCREF(value);
          PyDict_SetItemString(termios_nam2num,
            cc_cooked_vnames[i].name, value);
        }
      }
    }

    if (termios_num2nam == NULL) {
      i = l = 0;
      termios_num2nam = PyList_New(PyDict_Size(termios_nam2num));
      if (termios_num2nam == NULL) {
        return NULL;
      }
      while (PyDict_Next(termios_nam2num, &i, &key, &value)) {
        Py_INCREF(key);
        l = PyInt_AsLong(value);
        PyList_SetItem(termios_num2nam, l, key);
      }
    }

    /* Now we actually return a value */
    switch (value_type) {
      case GET_BAUDNAM: result = termios_baudnam; break;
      case GET_BAUDNUM: result = termios_baudnum; break;
      case GET_NAM2NUM: result = termios_nam2num; break;
      case GET_NUM2NAM: result = termios_num2nam; break;
      default:
        PyErr_SetString(PyExc_ValueError, "unknown value");
        return NULL;
    }
    if (result == NULL)
      PyErr_SetString(PyExc_SystemError, "cannot access TERMIOS values");
    else
      Py_INCREF(result);
    return result;
  }

static PyObject *
stty_termios(self, args)
  PyObject *self, *args;
  { PyObject *result = NULL;
    PyObject *termios_info, *cclist, *key, *item, *flag, *values;
    PyObject *iflag, *oflag, *cflag, *lflag, *ispeed, *ospeed;
    PyObject *termios_speed;
    struct termios *settings;
    int i, l, pos, ti_speed, to_speed;
    char *baudnam;

    settings = &Stty_settings(self);
    /* retrieve values from settings, return same value as
       termios.tcgetattr() */
    if (PyArg_ParseTuple(args, "")) {
      if (Stty_read_settings(self) == -1)
        return PyErr_SetFromErrno(PyExc_IOError);

      iflag  = get_stty_value(settings, "_iflag");
      if (iflag  == NULL) return NULL;
      oflag  = get_stty_value(settings, "_oflag");
      if (oflag  == NULL) return NULL;
      cflag  = get_stty_value(settings, "_cflag");
      if (cflag  == NULL) return NULL;
      lflag  = get_stty_value(settings, "_lflag");
      if (lflag  == NULL) return NULL;

      values = stty_termios_mod_info(settings, GET_BAUDNAM);
      if (values == NULL)
        return NULL;
      termios_get_speed(settings, &ti_speed, &to_speed);
      baudnam = termios_baud_name(ti_speed);
      ispeed = PyDict_GetItemString(values, baudnam);
      baudnam = termios_baud_name(to_speed);
      ospeed = PyDict_GetItemString(values, baudnam);
      Py_DECREF(values);

      flag   = get_stty_value(settings, "cc");
      if (flag   == NULL) return NULL;
      values = stty_termios_mod_info(settings, GET_NUM2NAM);
      if (values == NULL)
        return NULL;
      l = PyList_Size(values);
      cclist = PyList_New(l);
      if (cclist == NULL) return NULL;
      for (i = 0; i < l; i++) {
        key = PyList_GetItem(values, i);
        item = PyDict_GetItem(flag, key);
        Py_INCREF(item);
        PyList_SetItem(cclist, i, item);
      }
      Py_DECREF(flag);
      Py_DECREF(values);
      result = Py_BuildValue("NNNNNNN",
                             iflag, oflag, cflag, lflag,
                             ispeed, ospeed, cclist);

    /* set values to settings, give same value as termios.tcsetattr() */
    } else if (PyArg_ParseTuple(args,
                                "O!;termios expected optional list",
                                &PyList_Type, &termios_info)) {
      PyErr_Clear();
      if (PyArg_Parse(termios_info,
                      "O!O!O!O!O!O!O!:termios",
                      &PyInt_Type, &iflag,  &PyInt_Type, &oflag,
                      &PyInt_Type, &cflag,  &PyInt_Type, &lflag,
                      &PyInt_Type, &ispeed, &PyInt_Type, &ospeed,
                      &PyList_Type, &cclist)) {
        if (!set_stty_value(settings, "_iflag", iflag)) return NULL;
        if (!set_stty_value(settings, "_oflag", oflag)) return NULL;
        if (!set_stty_value(settings, "_cflag", cflag)) return NULL;
        if (!set_stty_value(settings, "_lflag", lflag)) return NULL;

        values = stty_termios_mod_info(settings, GET_BAUDNUM);
        if (values == NULL)
          return NULL;
        termios_speed = PyDict_GetItem(values, ispeed);
        ti_speed = termios_baud_value(PyString_AsString(termios_speed));
        termios_speed = PyDict_GetItem(values, ospeed);
        to_speed = termios_baud_value(PyString_AsString(termios_speed));
        Py_DECREF(values);
        (void)termios_set_speed(settings, ti_speed, to_speed);

        values = stty_termios_mod_info(settings, GET_NUM2NAM);
        if (values == NULL)
          return NULL;
        for (i = 0; i < l; i++) {
          if ((item = PyList_GetItem(cclist, i)) == NULL)
            continue;
          if ((key = PyList_GetItem(values, i)) == NULL)
            continue;
          Py_INCREF(item);
          set_stty_value(settings, PyString_AsString(key), item);
        }
        Py_DECREF(values);

        if (Stty_write_settings(self, TCSANOW) == -1)
          PyErr_SetFromErrno(PyExc_IOError);
      }
      result = Py_None;
      Py_INCREF(Py_None);
    }
    return result;
  }

#endif /*STTY_TERMIOS_FUNCTION*/

static PyMethodDef stty_methods[] = {
#ifdef STTY_TERMIOS_FUNCTION
  { "termios", stty_termios, METH_VARARGS },
#endif /*STTY_TERMIOS_FUNCTION*/
  { NULL, NULL }
};

PyObject *
Stty_new(fd)
  int fd;
  { Stty *object;

    object = PyObject_NEW(Stty, &SttyType);
    if (object != NULL)
      Stty_fd(object) = fd;
    return (PyObject *)object;
  }

static void
Stty__dealloc(self)
  PyObject *self;
  {
    PyMem_DEL(self);
  }

static int
Stty__print(self, fp, flags)
  PyObject *self;
  FILE *fp;
  int flags;
  {
    fprintf(fp, "<stty fd=%d>", Stty_fd(self));
    return 0;
  }

static int
Stty__compare(self, other)
  PyObject *self, *other;
  { PyObject *f;
    int sfd, ofd;

    sfd = ofd = -1;

    f = NULL;
    if (Stty_Check(self))
      sfd = Stty_fd(self);
    else if (PyInt_Check(self))
#if PYTHON_API_VERSION >= 1007
      sfd = PyInt_AS_LONG(self);
#else
      sfd = PyInt_AsLong(self);
#endif

    else if (PyObject_HasAttrString(self, "fileno"))
      if ((f = PyObject_CallMethod(self, "fileno", NULL)) != NULL &&
          PyInt_Check(f))
#if PYTHON_API_VERSION >= 1007
        sfd = PyInt_AS_LONG(f);
#else
        sfd = PyInt_AsLong(f);
#endif

    else if ((f = PyNumber_Int(self)) != NULL)
#if PYTHON_API_VERSION >= 1007
      sfd = PyInt_AS_LONG(f);
#else
      sfd = PyInt_AsLong(f);
#endif
    Py_XDECREF(f);

    f = NULL;
    if (Stty_Check(other))
      ofd = Stty_fd(other);
    else if (PyInt_Check(other))
#if PYTHON_API_VERSION >= 1007
      ofd = PyInt_AS_LONG(other);
#else
      ofd = PyInt_AsLong(other);
#endif

    else if (PyObject_HasAttrString(other, "fileno"))
      if ((f = PyObject_CallMethod(other, "fileno", NULL)) != NULL &&
          PyInt_Check(f))
#if PYTHON_API_VERSION >= 1007
        ofd = PyInt_AS_LONG(f);
#else
        ofd = PyInt_AsLong(f);
#endif

    else if ((f = PyNumber_Int(other)) != NULL)
#if PYTHON_API_VERSION >= 1007
      ofd = PyInt_AS_LONG(f);
#else
      ofd = PyInt_AsLong(f);
#endif
    Py_XDECREF(f);
    if (sfd == -1 || ofd == -1)
      PyErr_SetString(PyExc_TypeError, "expecting a file object");
    else
      PyErr_Clear();
    return (sfd < ofd ? -1 : sfd > ofd ? 1 : 0);
  }

static PyObject *
Stty__getattr(self, attrname)
  PyObject *self;
  char *attrname;
  { PyObject *result = NULL;
    struct termios *settings;

    if (Stty_read_settings(self) == -1)
      return PyErr_SetFromErrno(PyExc_IOError);

    settings = &Stty_settings(self);

    if (strcmp(attrname, "_fd") == 0)
      result = PyInt_FromLong(Stty_fd(self));
    else if (strcmp(attrname, "__members__") == 0)
      result = Py_BuildValue("[sssssssssss]",
        "_cflag", "_fd", "_iflag", "_lflag", "_oflag",
        "baud", "cc", "cooked", "echo", "rates", "raw"
      );

    else if (strcmp(attrname, "rates") == 0)
      result = termios_baud_rates();
    else if (strcmp(attrname, "_cflag") == 0 ||
             strcmp(attrname, "_iflag") == 0 ||
             strcmp(attrname, "_lflag") == 0 ||
             strcmp(attrname, "_oflag") == 0)
      result = get_stty_value(settings, attrname);
    else {
      result = get_stty_value(settings, attrname);
      if (result == NULL &&
          PyErr_Occurred() &&
          PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
        result = Py_FindMethod(stty_methods, self, attrname);
      }
    }

    return result;
  }

static int
Stty__setattr(self, attrname, value)
  PyObject *self, *value;
  char *attrname;
  { int rc = -1;
    char error_msg[256];
    int i, intvalue;
    PyObject *f;
    struct termios *settings;

    if (Stty_read_settings(self) == -1) {
      PyErr_SetFromErrno(PyExc_IOError);
      return -1;
    }

    if (value == NULL) {
      sprintf(error_msg, "cannot delete %s", attrname);
      PyErr_SetString(PyExc_ValueError, error_msg);
      return -1;
    }

    settings = &Stty_settings(self);
    if (strcmp(attrname, "baud") == 0) {
      rc = set_stty_value(settings, "baud", value) ? 0 : -1;
    } else if (strcmp(attrname, "cc") == 0) {
      if (!PyDict_Check(value))
        PyErr_SetString(PyExc_TypeError, "expecting dictionary");
      else
        rc = set_stty_value(settings, attrname, value) ? 0 : -1;
    } else if (strcmp(attrname, "cooked") == 0) {
      rc = set_stty_value(settings, "cooked", value) ? 0 : -1;
    } else if (strcmp(attrname, "echo") == 0) {
      rc = set_stty_value(settings, "echo", value) ? 0 : -1;
    } else if (strcmp(attrname, "raw") == 0) {
      rc = set_stty_value(settings, "raw", value) ? 0 : -1;
    } else if (strcmp(attrname, "_cflag") == 0 ||
               strcmp(attrname, "_iflag") == 0 ||
               strcmp(attrname, "_lflag") == 0 ||
               strcmp(attrname, "_oflag") == 0) {
      rc = set_stty_value(settings, attrname, value) ? 0 : -1;
    }
    if (rc != -1)
      if ((rc = Stty_write_settings(self, TCSANOW)) == -1)
        PyErr_SetFromErrno(PyExc_IOError);
    return rc;
  }

static PyObject *
Stty__call(self, args, keywds)
  PyObject *self, *args, *keywds;
  { PyObject *result = NULL;
    PyObject *s, *t, *u;
    int rc;
    int t_sz, d_sz;
    int t_rc, d_rc;

    /* are there arguments? keywds might be NULL */
    t_sz = PyTuple_Size(args);
    d_sz = (keywds != NULL) ? PyDict_Size(keywds) : 0;
    if (t_sz > 0 || d_sz > 0) {
      PyErr_Clear();

      if (Stty_read_settings(self) == -1)
        return PyErr_SetFromErrno(PyExc_IOError);

      t = u = NULL;
      /* there is only one argument and it is a tuple or dictionary */
      if (t_sz == 1 && d_sz == 0) {
#if PYTHON_API_VERSION >= 1007
        s = PyTuple_GET_ITEM(args, 0);
#else
        s = PyTuple_GetItem(args, 0);
#endif
        if (PyTuple_Check(s))
          t = s;
        else if (PyDict_Check(s))
          u = s;
      /* there are two arguments and is a tuple and dictionary, in order */
      } else if (t_sz == 2 && d_sz == 0) {
#if PYTHON_API_VERSION >= 1007
        t = PyTuple_GET_ITEM(args, 0);
#else
        t = PyTuple_GetItem(args, 0);
#endif
#if PYTHON_API_VERSION >= 1007
        u = PyTuple_GET_ITEM(args, 1);
#else
        u = PyTuple_GetItem(args, 1);
#endif
        if (!PyTuple_Check(t) || !PyDict_Check(u))
          t = u = NULL;
      }

      if (t == NULL && u == NULL) {
        t = args;
        u = keywds;
      }
      t_rc = (t != NULL) ? Stty_parse_stty_list(self, t) : 1;
      d_rc = (u != NULL) ? Stty_parse_stty_dict(self, u) : 1;
      if (!t_rc || !d_rc)
        ;
      else if (Stty_write_settings(self, TCSANOW) == -1)
        return PyErr_SetFromErrno(PyExc_IOError);
      else {
        result = Py_None;
        Py_INCREF(result);
      }

    /* no arguments */
    } else
#ifdef STTY_RETURNS_DICT
      result = Stty_unparse_stty_dict(self);
#else
      result = Stty_unparse_stty_tuple(self);
#endif
    return result;
  }

static char Stty__doc__[] =
  "A stty object, to manage termios data";

static PyTypeObject SttyType = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "stty",
  sizeof(Stty),
  0,
  Stty__dealloc,            /* tp_dealloc */
  Stty__print,              /* tp_print */
  Stty__getattr,            /* tp_getattr */
  Stty__setattr,            /* tp_setattr */
  Stty__compare,            /* tp_compare */
  0,                        /* tp_repr */
  0,                        /* tp_as_number */
  0,                        /* tp_as_sequence */
  0,                        /* tp_as_mapping */
  0,                        /* tp_hash */
  Stty__call,               /* tp_call */
  0,                        /* tp_str */
  0,                        /* tp_getattro */
  0,                        /* tp_setattro */
  0,                        /* tp_as_buffer */
  0,                        /* tp_xxx4 */
  Stty__doc__,              /* tp_doc */
};

static PyObject *
stty(unused, args)
  PyObject *unused, *args;
  { PyObject *newobj;
    PyObject *value, *v;
    int fd;

    if (!PyArg_ParseTuple(args, "O", &value))
      return NULL;
    if (PyInt_Check(value)) {
#if PYTHON_API_VERSIO >= 1007
      fd = PyInt_AS_LONG(value);
#else
      fd = PyInt_AsLong(value);
#endif
    } else if (PyObject_HasAttrString(value, "fileno")) {
      if ((v = PyObject_CallMethod(value, "fileno", NULL)) == NULL)
        return NULL;
      if (!PyInt_Check(value)) {
        PyErr_SetString(PyExc_ValueError, "expected integer");
        return NULL;
      }
#if PYTHON_API_VERSION >= 1007
      fd = PyInt_AS_LONG(v);
#else
      fd = PyInt_AsLong(v);
#endif
    } else {
      PyErr_SetString(PyExc_TypeError, "expected integer or file object");
      return NULL;
    }
    newobj = Stty_new(fd);
    return newobj;
  }

static PyMethodDef stty_Methods[] = {
  { "stty", stty, METH_VARARGS },
  { NULL, NULL },
};

void
initstty()
  {
    (void)Py_InitModule("stty", stty_Methods);
  }
